import scipy as sp
from openTMM import Layer
import openTMM.tests.utils as utils

#---------------------------------------------------------------------------------------------------
def test(logfile, layerDict, freq, phiDeg):
    numFailedTests = 0
    numTests = 0

    # freq in Hz (range), phiDeg in Deg (range)
    freqSample = 50
    phiSample = 30
    xSample = 100
    numAvgLayersOutside = 0, 5

    freqSet = sp.linspace(freq[0], freq[1], freqSample)
    phiDegSet = sp.linspace(phiDeg[0], phiDeg[1], phiSample)

    for patholMethod in 'keepReal', 'makeComplex':
        layerStack = Layer(layerDict, patholMethod)

        for pol in "parallel", "perp":
            for f in freqSet:
                for phi in phiDegSet:
                    for num in numAvgLayersOutside:
                        numTests += 1
                        omega = 2.0 * sp.pi *f
                        x, u_e, u_m = layerStack.energy(f, phi, xSample, pol, num)
                        x, Q_e, Q_m = layerStack.loss(f, phi, xSample, pol, num)
                        x, divS = layerStack.divPoynting(f, phi, xSample, pol, num)

                        # avoid subtractive cancellation
                        divSscaled = divS / omega
                        LHSimag = divSscaled.imag + 2.0 * u_e
                        RHSimag = 2.0 * u_m
                        ix = divSscaled.imag < 0
                        LHSimag[ix] = divSscaled[ix].imag - 2.0 * u_m[ix]
                        RHSimag[ix] = -2.0 * u_e[ix]
                        
                        testRealPassed = utils.assertApproxEqual(divSscaled.real, -(Q_e+Q_m)/omega)
                        testImagPassed = utils.assertApproxEqual(LHSimag, RHSimag)
                        if not (testRealPassed and testImagPassed):
                            data = {'omega':omega, 'phi':phi} 
                            data['u'] = {}
                            data['u']['u_e'] = u_e
                            data['u']['u_m'] = u_m
                            data['Q'] = {}
                            data['Q']['Q_e'] = Q_e
                            data['Q']['Q_m'] = Q_m
                            data['divS'] = divS
                            
                            
                            msg = 'real part test ' + str(testRealPassed)
                            msg += '.  imag. part test ' + str(testImagPassed)
                            dataSet = logfile.minDataSet(layerStack, layerDict, data, 
                                                         'PoyntingThm', msg)
                            logfile.write(dataSet)
                            numFailedTests
    return numFailedTests, numTests
#---------------------------------------------------------------------------------------------------
def run(logfile):

    def printTestResult(sampleNum, numFailedTests, numTests):
        msg = 'Poynting Thm. test results for sample ' + str(sampleNum) + ' : '
        if numFailedTests == 0:
            msg = msg + 'Passed ' + str(numTests) + ' Tests'
        elif numFailedTests != 0:
            msg = msg + 'Failed ' + str(numFailedTests) + ' of ' + str(numTests) + ' Tests' 
        print msg

    def printTestStatus(sampleNum):
        print 'Performing Poynting Thm. test on sample ' + str(sampleNum) + ' of 5.'
            

    mm = 1.0*10**-3
    GHz = 1.0*10**9

    freqRange = sp.array([100.0, 200.0], float) * GHz
    phiRange = sp.array([0, 89.0], float)

    # 1st Sample, RHM without absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon''_{\ell} = 0, \mu''_{\ell} \} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 1
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0, 9.0], float)
    muRange = sp.array([1.0, 9.0], float)
    hRange = sp.array([1, 10], float) * mm

    layerDict = utils.randomIdealLayers(epsRange, muRange, hRange, 25)
    fNum, total = test(logfile, layerDict, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 2nd Sample, RHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon''_{\ell} \in \mathbb{R^+} , \mu''_{\ell} \} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 2
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0+.001J, 9.0 + 1.0J], complex)
    muRange = sp.array([1.0, 9.0], float)
    hRange = sp.array([1, 10], float) * mm

    layerDict = utils.randomLayers(epsRange, muRange, hRange, 25)
    fNum, total = test(logfile, layerDict, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 3rd Sample, RHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon''_{\ell} = 0 , \mu''_{\ell} \} \in \mathbb{R^+} $
    #-----------------------------------------------------------------------------------------------
    sampleNum = 3
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0, 15.0], float)
    muRange = sp.array([1.0 + .01J, 9.0 + 0.1J], complex)
    hRange = sp.array([1, 10], float) * mm

    layerDict = utils.randomLayers(epsRange, muRange, hRange, 50)
    fNum, total = test(logfile, layerDict, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 4th Sample, RHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon''_{\ell} \in \mathbb{R^+} , \mu''_{\ell} \} \in \mathbb{R^+} $
    #-----------------------------------------------------------------------------------------------
    sampleNum = 4
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0 + .01J, 15.0 + 0.5J ], complex)
    muRange = sp.array([1.0 + .01J, 9.0 + 0.1J], complex)
    hRange = sp.array([1, 10], float) * mm

    layerDict = utils.randomLayers(epsRange, muRange, hRange, 50)
    fNum, total = test(logfile, layerDict, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 5th Sample, RHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon''_{\ell} = \in \mathbb{R^+} , \mu''_{\ell} \} \in \mathbb{R^+} $
    # 0th layer is absorbant and pth layer is dielectric.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 5
    printTestStatus(sampleNum)
    epsRangeIdeal = sp.array([1.0, 15.0], float)
    muRangeIdeal = sp.array([1.0, 9.0], float)
    hRangeIdeal = sp.array([1, 10], float) * mm

    epsRange = sp.array([0.1 + 0.001J, 4.0 + 0.1J], complex)
    muRange = sp.array([0.1 + 0.01, 8.0 + 0.05J], complex)
    hRange = sp.array([1, 5], float) * mm 

    layerDictIdeal = utils.randomIdealLayers(epsRangeIdeal, muRangeIdeal, hRangeIdeal, 25)
    layerDictAbsorp = utils.randomLayers(epsRange, muRange, hRange, 25)

    layerDict = utils.shuffleLayers(layerDictIdeal, layerDictAbsorp)
    layerDict['hostMedia'] = {}
    layerDict['hostMedia']['epsilonRelative'] = 2.0 * sp.ones(2, complex)
    layerDict['hostMedia']['muRelative'] = 3.0 * sp.ones(2, complex)    
    layerDict['hostMedia']['epsilonRelative'][0] = 2.5 + 0.5J
    layerDict['hostMedia']['muRelative'][0] = 1.5 + 0.1J

    fNum, total = test(logfile, layerDict, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)

#---------------------------------------------------------------------------------------------------



'''
Provides Boundary and Layer classes.  To access them use:
  import openTMM as tmm
  tmm.Boundary # Boundary class
  tmm.Layer # layer class

Boundary class -- Performs low-level calculations associated with the transfer matrix method (TMM).
Layer class -- Performs high-level calculations such as computing the time-averaged 
               electric/magnetic energy density, the transverse component of the electric field,
               and the transmission and reflection coefficients.

For details, see Table 1 and 2 in A. J. Yuffa, J. A. Scales, Object-oriented electrodynamic transfer
matrix code with modern applications,  preprint submitted to Computer Physics Communications (CPC).
The preprint is distributed with the source code, see openTMMpreprint.pdf

If you want to access the modules where the two classes are implemented you must explicitly 
import them, i.e. :
  from openTMM.mods import boundary, layer

To test the two classes as described in the paper use:
  from openTMM.tests import main
  main.run()
'''
__version__ = '0.0.1'
from mods.boundary import Boundary
from mods.layer import Layer
del mods




import openTMM.tests.logFile as LF
import openTMM.tests.testReciprocity as testReciprocity
import openTMM.tests.testEnergy as testEnergy
import openTMM.tests.testFIM as testFIM
import openTMM.tests.testWeston as testWeston

def run():
    print "Testing will take about 45 minutes.\n"
    print "Coffee time?  \n\n"
    logfile = LF.LogFile()

    testReciprocity.run(logfile)
    print '--------------------------------\n\n'
    testEnergy.run(logfile)
    print '--------------------------------\n\n'
    testFIM.run(logfile)
    print '--------------------------------\n\n'
    testWeston.run(logfile)
    print '--------------------------------\n\n'

    if logfile.saveFile:
        msg = "At least one test failed.  If the number of the failed tests \n"
        msg += "is 'small' then everything should be fine.  Some tests \n"
        msg += "may have failed due to subtraction cancellation or \n"
        msg += "numerical swamping and there is no reason to be concerned.\n\n"

        msg += "If you feel that the number of the failed tests exceeds a 'reasonable'\n"
        msg += "amount then please email " + str(logfile.filename) + " to \n"
        msg += "ayuffa@gmail.com.  The file may be large so please zip it: \n"
        msg += "   $ gzip " + str(logfile.filename) + "\n"
        msg += "Thank you, \n    Alex J. Yuffa"

        print msg

    logfile.cleanUp()

if __name__ == '__main__':
    run()



import scipy as sp
from scipy import constants
from fortran_utils import get_chi, cmplx_sqrt

class Boundary:
    '''
    Boundary class -- Performs low-level calculations associated with the transfer matrix method 
                      (TMM).  This class is meant to be a base class (super-class in the Python 
                      lexicon) that will be inherited by the derived classes (subclasses in the 
                      Python lexicon).  This class computes and sets the following quantities:

                      self.h - thickness of each layer
                      self.epsRel - relative permittivity
                      self.muRel - relative permeability
                      self.pol - polarization state
                      self.kx - x-component of the wavevector, $k_{x,\ell}$
                      self.chiPlus - transverse component of the electric field evaluated
                                     on the interface, $\chi^+_{\ell} / \chi^+_p$
                      self.chiMinus - transverse component of the electric field evaluated 
                                      on the interface, $\chi^-_{\ell} / \chi^+_p$

                      See Table 1 in A. J. Yuffa, J. A. Scales, Object-oriented electrodynamic 
                      transfer matrix code with modern applications,  preprint submitted to Computer
                      Physics Communications (CPC).  The preprint is distributed with the source 
                      code, see openTMMpreprint.pdf

                      USAGE:  After creating a Boundary object via __init__ e.g.,
                                bdStack = openTMM.Boundary(stackDict)
                              all quantities described above are computed and ready to be used, i.e,
                                bdStack.xxx 
                              where xxx is h, epsRel, muRel, pol, kx, chiPlus or chiMinus
    '''

# For best viewing, set your editors' column width to 100, the usual default value is 80.
#---------------------------------------------------------------------------------------------------
    def __init__(self, stack, patholMethod='makeComplex'):
        '''
        INPUT Type: 
        stack -- dictionary
          stack['height'] -- 1D, ndarray, float, e.g., sp.array([1,2,3], float)
          stack['epsilonRelative'] -- 1D, ndarray, float/complex, e.g.,  
                                      sp.array([1+1J,3,4J], complex)
          stack['muRelative'] -- 1D, ndarray, float/complex
          stack['hostMedia']['epsilonRelative'] -- 1D, ndarray, float/complex
          stack['hostMedia']['muRelative'] -- 1D, ndarray float/complex

        patholMethod -- string

        INPUT DESCRIPTION:  
        stack - mandatory [see Fig. 1 in the paper]
          Let arg = stack['xxx'], where 'xxx' = 'height' or 'epsilonRelative' or 'muRelative'.
          arg[0] refers to the 1st layer, $\ell=1$
                 NOTE: $\ell=1$ layer is next to the host medium of the transmitted field.
          arg[1] refers to the 2nd layer, $\ell=2$
          arg[n] refers to the nth layer, $\ell=n < p$, where $p^{th}$ region is the host medium
                 of the reflected field.

        stack['hostMedia'] - optional [see Fig. 1 in the paper]
          Let arg = stack['hostMedia']['xxx'], where 'xxx' = 'epsilonRelative' or 'muRelative'.  
          If stack['hostMedia'] is not given then the host media is assumed to be free space.
          arg[0] refers to the host medium for the transmitted field, $\ell=0$
          arg[1] refers to the host medium for the incident field, $\ell=p$

        patholMethod = 'makeComplex' (default) or 'keepReal' [see Sec. 3.2 in the paper]
          If 'makeComplex' then small absorption is added to the zero absorption layers.
          If 'keepReal' then no absorption is added to the zero absorption layers and
          the code may produce erroneous results at or very near the critical angles.
          Note that patholMethod only effects the pathological cases, i.e., 
          when $ \left\{ \epsilon_{\ell}, \mu_{\ell} \right\} \in \mathbb{R} $ and
          $ \epsilon_{\ell} \mu_{\ell} \omega^2 \geq k_{y,p}^2 $ [see Sec. 3.1 in the paper]
        '''
        self.h = None
        self.__setHeight(stack)
        self.epsRel = None
        self.__setEpsilonRel(stack)
        self.muRel = None
        self.__setMuRel(stack)


        # Zero absorption layers cause all of the pathological cases and must be identified
        # and dealt with according to the patholMethod (pathological method) selected.
        # [see Sec. 3.2 in the paper]
        self._ixIdealLHM = None
        self._ixIdealRHM = None
        self.__idZeroAbsorptionLayers()
        self._patholMethod = patholMethod
        if patholMethod == 'makeComplex':
            self.__fixZeroAbsorptionLayers()

        # The following are set by __setEMparams during $\chi^{\pm}_{\ell}$ computation.
        self.pol = None
        self.kx = None
        self.w = None

        # The following are set by setChi.
        self.chiPlus = None
        self.chiMinus = None
#---------------------------------------------------------------------------------------------------
    def __setHeight(self, stack):
        '''
        DESCRIPTION: 
        This function stores the height of each layer into self.h[ell], $\ell=0,\ldots,p$.

        INPUT:
        stack as described in __init__
        
        OUTPUT: Nothing is returned but self.h is stored.
        Let h = stack['height'] then self.h contains the height of each layer and host media.  
        The height of host medium is taken to be zero, i.e.,
        self.h[0] = 0.0 -- 'transmitted' half-space, $\ell=p$ 
        self.h[1] -- height of the 1st layer, $\ell=1$.
        self.h[2] -- height of the 2nd layer, $\ell=2$.
        self.h[h.size+1] -- height of the $\ell=p-1$ layer. 
        self.h[h.size+2] = 0.0 -- 'reflected' half-space, $\ell=0$.
        Thus, self.h[ell] is defined for $\ell = 0, \ldots, p$.
        [see Fig. 1 and Sec. 3.3.1 in the paper]
        '''
        h = stack['height']
        self.h = sp.zeros(h.size+2, float)
        self.h[1:h.size+1] = h
#---------------------------------------------------------------------------------------------------
    def __setEpsilonRel(self, stack):
        '''
        DESCRIPTION: 
        This function stores relative permittivity of each layer and host media into self.epsRel

        INPUT:
        stack as described in __init__

        OUTPUT: Nothing is returned but self.epsRel is stored.
        Let epsRel = stack['epsilonRelative'] then self.epsRel contains the relative permittivity
        of each layer and host media, i.e.,
        self.epsRel[0] -- relative permittivity of the 'transmitted' half-space, $\ell=0$.
        self.epsRel[1] -- relative permittivity of the 1st layer, $\ell=1$.
        self.epsRel[epsRel.size+1] -- relative permittivity of the $\ell=p-1$ layer.
        self.epsRel[epsRel.size+2] -- relative permittivity of the 'reflected' half-space, $\ell=0$.
        [see Fig. 1 and Sec. 3 in the paper]
        '''
        epsRel = stack['epsilonRelative']
        self.epsRel = sp.ones(epsRel.size+2, complex)
        self.epsRel[1:epsRel.size+1] = epsRel

        if stack.has_key('hostMedia'):
            # probably should check that stack['hostMedia']['epsilonRelative'][1] is real
            self.epsRel[0] = stack['hostMedia']['epsilonRelative'][0]
            self.epsRel[-1] = stack['hostMedia']['epsilonRelative'][1]
#---------------------------------------------------------------------------------------------------
    def __setMuRel(self, stack):
        '''
        DESCRIPTION: 
        This function stores relative permeability of each layer and host media into self.muRel

        INPUT:
        stack as described in __init__

        OUTPUT: Nothing is returned but self.muRel is stored.
        Let muRel = stack['muRelative'] then self.muRel contains the relative permeability
        of each layer and host media, i.e.,
        self.muRel[0] -- relative permeability of the 'transmitted' half-space, $\ell=0$.
        self.muRel[1] -- relative permeability of the 1st layer, $\ell=1$.
        self.muRel[muRel.size+1] -- relative permeability of the $\ell=p-1$ layer.
        self.muRel[muRel.size+2] -- relative permeability of the 'reflected' half-space, $\ell=0$.
        [see Fig. 1 and Sec. 3 in the paper]
        '''
        muRel = stack['muRelative']
        self.muRel = sp.ones(muRel.size+2, complex)
        self.muRel[1:muRel.size+1] = muRel

        if stack.has_key('hostMedia'):
            # probably should check that stack['hostMedia']['muRelative'][1] is real
            self.muRel[0] = stack['hostMedia']['muRelative'][0]
            self.muRel[-1] = stack['hostMedia']['muRelative'][1]
#---------------------------------------------------------------------------------------------------
    def __idZeroAbsorptionLayers(self):
        '''
        DESCRIPTION:
        This function uses self.epsRel and self.muRel to identify zero absorption layers.
        The identification of zero absorption layers is needed because zero absorption layers
        cause all pathological cases.  [see Sec. 3.1 in the paper]

        INPUT: None.  The function uses self.epsRel and self.muRel

        OUTPUT: Nothing is returned but self._ixIdealRHM and self._ixIdealLHM are stored.
        self._ixIdealRHM[ell] is set to True if $\ell^{th}$ layer is made of a right-handed material
        with zero absorption and self._ixIdealRHM[ell] is set to False otherwise, $\ell=0,\ldots,p$.
        self._ixIdealLHM[ell] is set to True if $\ell^{th}$ layer is made of a left-handed material
        with zero absorption and self._ixIdealLHM[ell] is set to False otherwise, $\ell=0,\ldots,p$.
        '''
        epsRel = self.epsRel
        muRel = self.muRel

        # Recall that according to the IEEE 754 standards, zero is represented exactly,
        # e.g., see http://steve.hollasch.net/cgindex/coding/ieeefloat.html or
        # http://en.wikipedia.org/wiki/Floating_point
        # Thus, hard comparison should be fine, assuming that the relative permittivity and 
        # relative permeability used to initialize the Boundary class did NOT come from some 
        # some previous calculation.
        ixZeroAbsorption = sp.isreal(epsRel) & sp.isreal(muRel)
        self._ixIdealRHM = ixZeroAbsorption & ~sp.signbit(epsRel.real) & ~sp.signbit(muRel.real)
        self._ixIdealLHM = ixZeroAbsorption & sp.signbit(epsRel.real) & sp.signbit(muRel.real)
#---------------------------------------------------------------------------------------------------
    def __fixZeroAbsorptionLayers(self):
        '''
        DESCRIPTION:
        This function adds a small imaginary part to absorption free layers identified by
        self._ixIdealRHM and self._ixIdealLHM.  Note $\ell=p$ layer is NOT modified because
        this is 'reflected' half-space which MUST be absorption free.  [see Sec. 3 in the paper]

        INPUT: None. The function uses self.epsRel, self.muRel

        OUTPUT: Nothing is returned but self.epsRel and self.muRel are modified for layers
                where self._ixIdealRHM or self._ixIdealLHM is True
        '''
        ixZeroAbsorption = self._ixIdealRHM | self._ixIdealLHM

        small = 10**-100
        self.epsRel[ixZeroAbsorption] += small * 1.0J
        self.muRel[ixZeroAbsorption] += small * 1.0J
        # $\ell=p$ layer must be kept with zero absorption, because the wave 
        # is incident from the $p^{th}$ layer [see Sec. 3 in the paper] 
        self.epsRel[-1] = self.epsRel[-1].real
        self.muRel[-1] = self.muRel[-1].real
#---------------------------------------------------------------------------------------------------
    def __setEMparams(self, f, phi, pol):
        '''
        DESCRIPTION:
        This function computes and stores $k_{x,\ell=0,\ldots,p}$ and $w_{\ell=0,\ldots,p}$ 
        into self.  Polarization state is also stored into self, i.e., self.pol
        [see Chapter 3 in the paper]

        INPUT:
        f - frequency in Hertz
          - float, scalar
        phi - angle of incidence in radians
            - float, scalar
        pol - polarization, 'parallel' if the electric field is parallel to the $xy$ plane or 'perp'
              if the electric field is perpendicular to the $xy$-plane. [see Fig.1 in the paper]
            - string
        
        OUTPUT: Nothing is returned but self.kx, self.w and self.pol are stored.
        self.pol - polarization state, 'parallel' or 'perp'
        self.kx[ell] - x-component of the wavevector, i.e, $k_{x,\ell}$ for $\ell=0,\ldots,p$. 
        self.w[ell] - scaled version of self.kx[ell], scaling depends on polarization state.
                      [see (16) and (22) in the paper]
        '''
        self.pol = pol
        epsRel = self.epsRel
        muRel = self.muRel
        omega = 2.0 * sp.pi * f

        # Compute scaled $k_{x, \ell = 0,\ldots,p}$ and choose physically appropriate root, i.e.,
        # imaginary part of $k_{x, \ell = 0,\ldots,p} > 0$, [see (8) in the paper]
        kypSquared = epsRel[-1].real * muRel[-1].real * sp.sin(phi)**2
        kx = cmplx_sqrt(epsRel * muRel - kypSquared)

        # handle pathological cases with zero absorption according to 'keepReal' scheme.
        # [see second scheme in Sec. 3.2 of the paper]
        if self._patholMethod == 'keepReal':
            ixRealKx = sp.isreal(kx)

            ixLHM = self._ixIdealLHM & ixRealKx
            kx[ixLHM] = -sp.absolute(kx[ixLHM].real)

            ixRHM = self._ixIdealRHM & ixRealKx
            kx[ixRHM] = sp.absolute(kx[ixRHM].real)

        # Fix $k_{x,p}$ if $p^{th}$ layer is made of a left-handed material because $p^{th}$ layer 
        # is always absorption free.  Moreover, note that __fixZeroAbsorptionLayers 
        # did NOT add a small imaginary part to self.epsRel[p] and self.muRel[p]
        if self._patholMethod == 'makeComplex' and self._ixIdealLHM[-1]:
            kx[-1] = -sp.absolute(kx[-1].real)

        # compute scaled version of $w_{\ell=0,\ldots,p}$ parameter [see (16) and (22) in the paper]
        if self.pol == 'parallel':
            w = epsRel / kx
        elif self.pol == 'perp':
            w = -kx / muRel

        # Un-scale kx and w (SI unit system)
        self.kx = kx * omega / constants.c
        self.w = w * constants.epsilon_0 * constants.c
#---------------------------------------------------------------------------------------------------
    def setChi(self, f, phi, pol):
        '''
        DESCRIPTION:
        This function computes and stores $\chi^+_{\ell=0,\ldots,p}$ and $\chi^-_{\ell=0,\ldots,p}$ 
        into self.  Note: $\chi^+_{\ell=p}$ is set to one, thus, $\chi^{\pm}_{\ell=0,\ldots,p}$ are
        really $\frac{\chi^{\pm}_{\ell=0,\ldots,p}}{\chi^+_p}$. $\chi^-_{\ell=0}$ is set to zero
        because there is no reflected wave in the $0^{th}$ region.
        [see Fig. 1 and (14), (20) in the paper]

        INPUT:
        f - frequency in Hertz
          - float, scalar
        phi - angle of incidence in radians
            - float, scalar
        pol - polarization, 'parallel' if the electric field is parallel to the $xy$ plane or 'perp'
              if the electric field is perpendicular to the $xy$-plane. [see Fig.1 in the paper]
            - string
        
        OUTPUT: Nothing is returned but self.chiPlus and self.chiMinus are stored.
        self.chiPlus[ell] - $\chi^+_{\ell}$ for $\ell=0,\ldots,p$
        self.chiMinus[ell] - $chi^-_{\ell}$ for $\ell=0,\ldots,p$
        '''
        # Set all E&M parameters needed to compute $\chi^{\pm}_{\ell=0,\ldots,p}
        self.__setEMparams(f, phi, pol)

        self.chiPlus, self.chiMinus = get_chi(self.h, self.kx, self.w)
#---------------------------------------------------------------------------------------------------



import scipy as sp
import openTMM as tmm
import os

basePath = tmm.__path__[0]
curPath = os.path.abspath(os.curdir)

inFile = os.path.join(basePath, "tests", "cpcInput.txt")
outFileSVN = os.path.join(basePath, "tests", "cpcOutputSVN.txt")
outFile = os.path.join(curPath, "cpcOutput.txt")

# input data
inData = sp.loadtxt(inFile, delimiter=',', skiprows=1)

# Make Layer obj from inData
mm = 1.0 * 10**-3
GHz = 1.0 * 10**9
fo = inData[0,3] * GHz
phi = inData[0,4]
xSample = int(inData[0,5])
stackDict = {}
stackDict['epsilonRelative'] = inData[:,0]
stackDict['muRelative'] = inData[:,1]
stackDict['height'] = inData[:,2] * mm
stack = tmm.Layer(stackDict)

out = open(outFile, 'w')

# perp polarization
x, u_e, u_m = stack.energy(fo, phi, xSample, pol='perp')
out.write('Perpendicular polarization:\n')
out.write('u_e(x) = \n')
for i in range(u_e.size):
    s = "%*.*E  " % (1, 6, float(u_e[i]))
    out.write(s)

out.write('\n')
out.write('u_m(x) = \n')

for i in range(u_m.size):
    s = "%*.*E  " % (1, 6, float(u_m[i]))
    out.write(s)

# parallel polarization
x, u_e, u_m = stack.energy(fo, phi, xSample, pol='parallel')
out.write('\n\n')
out.write('Parallel polarization:\n')
out.write('u_e(x) = \n')

for i in range(u_e.size):
    s = "%*.*E  " % (1, 6, float(u_e[i]))
    out.write(s)

out.write('\n')
out.write('u_m(x) = \n')

for i in range(u_m.size):
    s = "%*.*E  " % (1, 6, float(u_m[i]))
    out.write(s)


out.close()
print '\n'
print 'Dear CPC Program Librarian,\n'
print 'INPUT DATA: ' + inFile
print '\nOUTPUT DATA (this run): ' + outFile
print '\nOUTPUT DATA (from SVN): ' + outFileSVN
print '\nPlease compare OUTPUT DATA (this run) with OUTPUT DATA (from SVN).'
print '\n\nThanks you, \n    Alex J. Yuffa <ayuffa@gmail.com>'





import scipy as sp
from openTMM import Layer
import openTMM.tests.utils as utils

def test(logfile, layerDict, freq):
    numFailedTests = 0
    numTests = 0

    # freq in Hz (range)
    freqSample = 300
    xSample = 100
    phi = 0.0
    numAvgLayersOutside = 0, 5

    freqSet = sp.linspace(freq[0], freq[1], freqSample)

    for patholMethod in 'keepReal', 'makeComplex':
        layerStack = Layer(layerDict, patholMethod)

        for pol in "parallel", "perp":
            for f in freqSet:
                numTests += 1
                layerStack.setChi(f, phi, pol)
                chiPlus = layerStack.chiPlus
                chiMinus = layerStack.chiMinus
                chi  = sp.absolute(chiPlus + chiMinus)
                chiPassed = utils.assertApproxEqual(chi, sp.absolute(chiPlus))

                if not chiPassed:
                    data = {'freq':f}
                    data['chi'] = {}
                    data['chiPlus'] = chiPlus
                    data['chiMinus'] = chiMinus
                    dataSet = logfile.minDataSet(layerStack, layerDict, data, 'westonChi')
                    logfile.write(dataSet)
                    numFailedTests += 1

                for num in numAvgLayersOutside:
                    numTests += 1
                    x, gammaPlus, gammaMinus = layerStack.field(f, phi, xSample, pol, num)
                    gamma = sp.absolute(gammaPlus + gammaMinus)
                    gammaPassed = utils.assertApproxEqual(gamma, sp.absolute(gammaPlus))
                    
                    if not gammaPassed:
                        data = {'freq':f} 
                        data['gamma'] = {}
                        data['gamma']['gammaPlus'] = gammaPlus
                        data['gamma']['gammaMinus'] = gammaMinus
                        dataSet = logfile.minDataSet(layerStack, layerDict, data, 'westonGamma')
                        logfile.write(dataSet)
                        numFailedTests += 1
    return numFailedTests, numTests
#---------------------------------------------------------------------------------------------------
def run(logfile):

    def printTestResult(sampleNum, numFailedTests, numTests):
        msg = 'Weston Thm. test results for sample ' + str(sampleNum) + ' : '
        if numFailedTests == 0:
            msg = msg + 'Passed ' + str(numTests) + ' Tests'
        elif numFailedTests != 0:
            msg = msg + 'Failed ' + str(numFailedTests) + ' of ' + str(numTests) + ' Tests' 
        print msg

    def printTestStatus(sampleNum):
        print 'Weston Thm. test on sample ' + str(sampleNum) + ' of 5.'
            

    mm = 1.0*10**-3
    GHz = 1.0*10**9

    freqRange = sp.array([50.0, 250.0], float) * GHz

    # 1st Sample, RHM with and without absorption.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 1
    printTestStatus(sampleNum)

    epsRangeIdeal = sp.array([0.01, 10.0], float)
    muRangeIdeal = sp.array([1.0, 1.0], float)
    hRangeIdeal = sp.array([1, 10], float) * mm
    layerDict1 = utils.randomIdealLayers(epsRangeIdeal, muRangeIdeal, hRangeIdeal, 25)

    epsRange = sp.array([0.01 + 0.001J, 10.0 + 1.0J], complex)
    muRange = sp.array([1.0, 1.0], complex)
    hRange = sp.array([1, 5], float) * mm
    layerDict2 = utils.randomLayers(epsRange, muRange, hRange, 25)

    layerDict = utils.shuffleLayers(layerDict1, layerDict2)
    layerDict['muRelative'] = layerDict['epsilonRelative'].copy()
    layerDict['hostMedia'] = {}
    layerDict['hostMedia']['epsilonRelative'] = 2.0 * sp.ones(2, complex)
    layerDict['hostMedia']['muRelative'] = 2.0 * sp.ones(2, complex)    


    fNum, total = test(logfile, layerDict, freqRange)
    printTestResult(sampleNum, fNum, total)


    # 2nd Sample, LHM with and without absorption.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 2
    printTestStatus(sampleNum)

    epsRangeIdeal = sp.array([-0.01, -10.0], float)
    muRangeIdeal = sp.array([-1.0, -1.0], float)
    hRangeIdeal = sp.array([1, 10], float) * mm
    layerDict1 = utils.randomIdealLayers(epsRangeIdeal, muRangeIdeal, hRangeIdeal, 25)

    epsRange = sp.array([-0.01 + 0.001J, -10.0 + 1.0J], complex)
    muRange = sp.array([-1.0, -1.0], complex)
    hRange = sp.array([1, 5], float) * mm
    layerDict2 = utils.randomLayers(epsRange, muRange, hRange, 25)

    layerDict = utils.shuffleLayers(layerDict1, layerDict2)
    layerDict['muRelative'] = layerDict['epsilonRelative'].copy()
    layerDict['hostMedia'] = {}
    layerDict['hostMedia']['epsilonRelative'] = -1.5 * sp.ones(2, complex)
    layerDict['hostMedia']['muRelative'] = -1.5 * sp.ones(2, complex)    


    fNum, total = test(logfile, layerDict, freqRange)
    printTestResult(sampleNum, fNum, total)


    # 3rd Sample, LHM and RHM with absorption.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 3
    printTestStatus(sampleNum)

    epsRangeLHM = sp.array([-0.01 + 0.01J, -10.0 + 0.1J], complex)
    muRangeLHM = sp.array([-1.0, -1.0], complex)
    hRangeLHM = sp.array([0.1, 7], float) * mm
    layerDictLHM = utils.randomLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 15)

    epsRangeRHM = sp.array([0.01 + 0.001J, 3.0 + 0.1J], complex)
    muRangeRHM = sp.array([1.0, 1.0], complex)
    hRangeRHM = sp.array([1, 10], float) * mm
    layerDictRHM = utils.randomLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 15)

    layerDict = utils.shuffleLayers(layerDictLHM, layerDictRHM)
    layerDict['muRelative'] = layerDict['epsilonRelative'].copy()
    layerDict['hostMedia'] = {}
    layerDict['hostMedia']['epsilonRelative'] = 3.0 * sp.ones(2, complex)
    layerDict['hostMedia']['muRelative'] = 3.0 * sp.ones(2, complex)    
    layerDict['hostMedia']['epsilonRelative'][0] = -2.5 + 0.01J
    layerDict['hostMedia']['muRelative'][0] = -2.5 + 0.01J

    fNum, total = test(logfile, layerDict, freqRange)
    printTestResult(sampleNum, fNum, total)


    # 4th Sample, LHM and RHM without absorption.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 4
    printTestStatus(sampleNum)

    epsRangeLHM = sp.array([-0.01, -5.0], float)
    muRangeLHM = sp.array([-0.1, -11.0], float)
    hRangeLHM = sp.array([1, 20], float) * mm
    layerDictLHM = utils.randomIdealLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 15)

    epsRangeRHM = sp.array([0.01, 15.0], float)
    muRangeRHM = sp.array([0.1, 10.0], float)
    hRangeRHM = sp.array([1, 10], float) * mm
    layerDictRHM = utils.randomIdealLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 15)

    layerDict = utils.shuffleLayers(layerDictLHM, layerDictRHM)
    layerDict['muRelative'] = layerDict['epsilonRelative'].copy()
    layerDict['hostMedia'] = {}
    layerDict['hostMedia']['epsilonRelative'] = -1.3 * sp.ones(2, float)
    layerDict['hostMedia']['muRelative'] = -1.3 * sp.ones(2, float)    
    layerDict['hostMedia']['epsilonRelative'][-1] = 2.5
    layerDict['hostMedia']['muRelative'][-1] = 2.5

    fNum, total = test(logfile, layerDict, freqRange)
    printTestResult(sampleNum, fNum, total)


    # 5th Sample, LHM and RHM with/without absorption.
    #-----------------------------------------------------------------------------------------------
    sampleNum = 5
    printTestStatus(sampleNum)

    epsRangeLHMideal = sp.array([-0.01, -10.0], float)
    muRangeLHMideal = sp.array([-0.1, -11.0], float)
    hRangeLHMideal = sp.array([1, 20], float) * mm
    layerDictLHMideal = utils.randomIdealLayers(epsRangeLHMideal, muRangeLHMideal, hRangeLHMideal, 5)

    epsRangeLHMabsorp = sp.array([-0.1 + 0.01J, -10.0 + 0.3J], complex)
    muRangeLHMabsorp = sp.array([-0.1, -11.0], float)
    hRangeLHMabsorp = sp.array([1, 8], float) * mm
    layerDictLHMabsorp = utils.randomLayers(epsRangeLHMabsorp, muRangeLHMabsorp, hRangeLHMabsorp, 5)

    layerDictLHM = utils.shuffleLayers(layerDictLHMideal, layerDictLHMabsorp)

    epsRangeRHMideal = sp.array([0.01, 15.0], float)
    muRangeRHMideal = sp.array([1.5, 5.0], float)
    hRangeRHMideal = sp.array([1, 10], float) * mm
    layerDictRHMideal = utils.randomIdealLayers(epsRangeRHMideal, muRangeRHMideal, hRangeRHMideal, 5)

    epsRangeRHMabsorp = sp.array([1.1, 10.0], float)
    muRangeRHMabsorp = sp.array([0.1 + 0.001J, -11.0 + 0.5J], complex)
    hRangeRHMabsorp = sp.array([1, 5], float) * mm
    layerDictRHMabsorp = utils.randomLayers(epsRangeRHMabsorp, muRangeRHMabsorp, hRangeRHMabsorp, 5)

    layerDictRHM = utils.shuffleLayers(layerDictRHMideal, layerDictRHMabsorp)

    layerDict = utils.shuffleLayers(layerDictLHM, layerDictRHM)
    layerDict['muRelative'] = layerDict['epsilonRelative'].copy()

    fNum, total = test(logfile, layerDict, freqRange)
    printTestResult(sampleNum, fNum, total)
#---------------------------------------------------------------------------------------------------



import scipy as sp
from scipy import constants
from boundary import Boundary as bd
from warnings import warn

LHMwarning = '\n\nThe physical meaning of the returned values of the called function \nmay be NOT what you have expected because the multilayer stack contains \nat least one layer made of a left-handed material (LHM).  The LHM must be \naccompanied by frequency dispersion, see the end of Sec. 6 in the paper.\n'

class Layer(bd):
    '''
    Layer class -- Performs high-level calculations such as computing the time-averaged 
                   electric/magnetic energy density, the transverse component of the electric field,
                   and the transmission and reflection coefficients.  The complete list of computed
                   quantities is given in Table 2 in A. J. Yuffa, J. A. Scales, Object-oriented 
                   electrodynamic transfer matrix code with modern applications,  preprint submitted
                   to Computer Physics Communications (CPC).  The preprint is distributed with the
                   source code, see openTMMpreprint.pdf
    '''
# Organization of the Layer class:
# The Layer class inherits the Boundary class and does NOT directly modify self.h, self.epsRel,
# self.muRel, self.pol, self.kx, self.w, self.chiPlus and self.chiMinus of the Boundary class.
# The public methods of the Layer class call the Boundary setChi method to compute self.kx, self.w,
# self.chiPlus and self.chiMinus and then calls an associated private method to convert them into
# a desired quantity.  For example, TRvsFreq calls Boundary.setChi first and then calls __T to 
# compute the transmission coefficient from the quantities computed by the Boundary.setChi method.

# For best viewing, set your editors' column width to 100, the usual default value is 80.
#---------------------------------------------------------------------------------------------------
    def __init__(self, stack, patholMethod='makeComplex'):
        '''
        INPUT Type: 
        stack -- dictionary
          stack['height'] -- 1D, ndarray, float, e.g., sp.array([1,2,3], float)
          stack['epsilonRelative'] -- 1D, ndarray, float/complex, e.g.,  
                                      sp.array([1+1J,3,4J], complex)
          stack['muRelative'] -- 1D, ndarray, float/complex
          stack['hostMedia']['epsilonRelative'] -- 1D, ndarray, float/complex
          stack['hostMedia']['muRelative'] -- 1D, ndarray float/complex

        patholMethod -- string

        INPUT DESCRIPTION:  
        stack - mandatory [see Fig. 1 in the paper]
          Let arg = stack['xxx'], where 'xxx' = 'height' or 'epsilonRelative' or 'muRelative'.
          arg[0] refers to the 1st layer, $\ell=1$
                 NOTE: $\ell=1$ layer is next to the host medium of the transmitted field.
          arg[1] refers to the 2nd layer, $\ell=2$
          arg[n] refers to the nth layer, $\ell=n < p$, where $p^{th}$ region is the host medium
                 of the reflected field.

        stack['hostMedia'] - optional [see Fig. 1 in the paper]
          Let arg = stack['hostMedia']['xxx'], where 'xxx' = 'epsilonRelative' or 'muRelative'.  
          If stack['hostMedia'] is not given then the host media is assumed to be free space.
          arg[0] refers to the host medium for the transmitted field, $\ell=0$
          arg[1] refers to the host medium for the incident field, $\ell=p$

        patholMethod = 'makeComplex' (default) or 'keepReal' [see Sec. 3.2 in the paper]
          If 'makeComplex' then small absorption is added to the zero absorption layers.
          If 'keepReal' then no absorption is added to the zero absorption layers and
          the code may produce erroneous results at or very near the critical angles.
          Note that patholMethod only effects the pathological cases, i.e., 
          when $ \left\{ \epsilon_{\ell}, \mu_{\ell} \right\} \in \mathbb{R} $ and
          $ \epsilon_{\ell} \mu_{\ell} \omega^2 \geq k_{y,p}^2 $ [see Sec. 3.1 in the paper]
        '''
        bd.__init__(self, stack, patholMethod)
#---------------------------------------------------------------------------------------------------
    def __R(self):
        '''
        DESCRIPTION: 
        This function computes the reflection coefficient.

        INPUT:  None
        __R uses self.chiMinus to compute the reflection coefficient
        
        OUTPUT:
        R -- reflection coefficient (real value).
        '''
        # [see (42b) and (43b) in the paper]
        chiMinus_p = self.chiMinus[-1]
        R = chiMinus_p * chiMinus_p.conj()
        return R.real
#---------------------------------------------------------------------------------------------------
    def __T(self):
        '''
        DESCRIPTION: 
        This function computes the transmission coefficient.

        INPUT:  None
        __T uses self to compute the transmission coefficient.

        OUTPUT:
        T -- transmission coefficient (real value).
        '''
        chiPlus_0 = self.chiPlus[0]
        eps_0 = self.epsRel[0] * constants.epsilon_0
        eps_p = self.epsRel[-1] * constants.epsilon_0
        mu_0 = self.muRel[0] * constants.mu_0
        mu_p = self.muRel[-1] * constants.mu_0
        kx_0 = self.kx[0]
        kx_p = self.kx[-1].real
        
        # [see (42a) and (43a) in the paper]
        if self.pol == 'perp':
            T = (mu_p / kx_p) * sp.real(kx_0.conj() / mu_0.conj()) * chiPlus_0 * chiPlus_0.conj()
        elif self.pol == 'parallel':
            T = kx_p * sp.real(eps_0.conj() * kx_0) / (eps_p * kx_0 * kx_0.conj()) * (
                chiPlus_0 * chiPlus_0.conj() )
        return T.real
#---------------------------------------------------------------------------------------------------
    def __field(self, xSample, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes transverse E-field; E_y if 'parallel' polarization or E_z if 'perp' polarization.
        [see (33) and (37) in the paper]

        INPUT:
        xSample - number of sample points per layer
        numAvgLayersOutside - How far (measured in average layer thickness) into 'reflected' 
                              and 'transmitted' half-spaces should the field by computed.

        OUTPUT:
        x[ell,:] - discretization of $\ell^{th}$ layer; the distance is measured from the
                   top of the layer.
        gammaPlus[ell,:] - transverse component of the E-field associated with the wave propagating 
                           in the positive x-direction, where $\ell$ denotes the layer number,
                           i.e., gammaPlus[ell,:] is evaluated at x[ell,:]
        gammaMinus[ell,:] - transverse component of the E-field associated with the wave propagating
                            in the negative x-direction, where $\ell$ denotes the layer number,
                            i.e., gammaMinus[ell,:] is evaluated at x[ell,:]

        Case 1:  If numAvgLayersOutside = 0 then __field 
        RETURNS x, gammaPlus, gammaMinus
        x, gammaPlus and gammaMinus contain half-spaces with all values set to zero.
        The reason for doing so, is to keep indexing the same as in the paper, i.e.,
        x[0,:], gammaPlus[0,:], gammaMinus[0,:] refers to $p^{th}$ region ('transmitted' half-space)
        x[-1,:], gammaPlus[-1,:], gammaMinus[-1,:] refers to $0^{th}$ region ('reflected' half-space)

        Case 2: If numAvgLayersOutside != 0 then __field 
        RETURNS x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside
        x, gammaPlus and gammaMinus are the same as in Case 1 and
        xOutside[0,:] - discretization of $0^{th}$ region ('transmitted' half-space); the
                        distance is measured from the top of the region, i.e., 1--0 interface. 
        gammaPlusOutside[0,:] - transverse component of the E-field in the $0^{th}$ region associated
                                with the wave propagating in the positive x-direction, i.e.,
                                gammaPlusOutside[0,:] is evaluated at xOutside[0,:]
        gammaMinusOutside[0,:] - transverse component of the E-field in the $0^{th}$ region
                                 associated with the wave propagating in the negative x-direction,
                                 i.e., gammaMinusOutside[0,:] is evaluated at xOutside[0,:]
        xOutside[-1,:] - discretization of $p^{th}$ region ('reflected' half-space); the distance
                         is measured from 'in-region' to the $p$--$p-1$ interface, i.e.,
                         xOutside[-1,-1] is the furthest point from the $p$--$p-1$ interface and
                         xOutside[-1,0] is the closest point to the $p$--$p-1$ interface
        gammaPlusOutside[-1,:] - transverse component of the E-field in the $p^{th}$ region 
                                 associated with the wave propagating in the positive x-direction, 
                                 i.e., gammaPlusOutside[-1,:] is evaluated at xOutside[-1,:]
        gammaMinusOutside[-1,:] - transverse component of the E-field in the $p^{th}$ region
                                  associated with the wave propagating in the negative x-direction,
                                  i.e., gammaMinusOutside[-1,:] is evaluated at xOutside[-1,:]
        '''
        chiPlus = self.chiPlus[:,sp.newaxis]
        chiMinus = self.chiMinus[:,sp.newaxis]
        kx = self.kx[:,sp.newaxis]
        h = self.h[:,sp.newaxis]

        # In what follows, i denotes the layer number, i.e., i = 0,...,p and 
        # j refers to the discretization of layer's thickness, i.e., j = 0,...,xSample-1
        # x contains the discretization for each layer, i.e,
        # x[i,0] is the TOP of i-th layer, i.e. boundary between i and i+1 layer.
        # x[i,-1] is the BOTTOM of i-th layer, i.e., boundary between i and i-1 layer
        # and x[0,:] = 0, x[-1,:] = 0 (half-spaces are set to zero) 
        # For example, if p=4, h=3 for each layer, and xSample=3, then
        # x[0,:] = [0.0, 0.0, 0.0]
        # x[i,:] = [0.0, 1.0, 2.0], for i = 1,2,3 [see Fig. 1 in the paper]
        # x[-1,:] = [0.0, 0.0, 0.0]

        x = sp.zeros((self.h.size, xSample), float)
        for ell in xrange(1, self.h.size-1):
            x[ell] = sp.linspace(0.0, self.h[ell], xSample, endpoint=False)

        gammaPlus = chiPlus * sp.exp(1.0J * kx * (x - h))
        gammaMinus = chiMinus * sp.exp(-1.0J * kx * (x - h))
        gammaPlus[0], gammaPlus[-1] = 0.0, 0.0
        gammaMinus[0], gammaMinus[-1] = 0.0, 0.0

        if numAvgLayersOutside == 0:
            return x, gammaPlus, gammaMinus

        elif numAvgLayersOutside > 0:
            num = numAvgLayersOutside
            avgH = h.mean()
            numPts = xSample * num
            xOutside = sp.zeros((2,numPts), float)
            gammaPlusOutside = sp.zeros(xOutside.shape, complex)
            gammaMinusOutside = sp.zeros(xOutside.shape, complex)

            # Computing transverse E-field in 'transmitted' half-space, i.e., $\ell=p$.
            xOutside[0] = sp.linspace(0.0, num*avgH, numPts, endpoint=False)
            gammaPlusOutside[0] = chiPlus[0] * sp.exp(1.0J * kx[0] * xOutside[0])

            # Computing transverse E-field in 'reflected' half-space, i.e., $\ell=0$
            xOutside[-1] = sp.linspace(-num*avgH, 0.0, numPts, endpoint=False)
            gammaPlusOutside[-1] = chiPlus[-1] * sp.exp(1.0J * kx[-1] * xOutside[-1])
            gammaMinusOutside[-1] = chiMinus[-1] * sp.exp(-1.0J * kx[-1] * xOutside[-1])

            return x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside
#---------------------------------------------------------------------------------------------------
    def __field2CoordSys(self, x, field, xOutside=None, fieldOutside=None):
        '''
        DESCRIPTION: 
        Converts field into a flat array as a function of distance as measured from the
        origin of the coordinate system set at the bottom of 'reflected' half-space, i.e., 
        $p$--$p-1$ interface.
        Note: Here field denotes any quantity that is defined inside all layers,
              e.g., field could be $u^{\text{(e,m)}}(x)$, $Q^{\text{(e,m)}}(x)$, etc.
        
        INPUT:
        x[ell,:] - discretization of $\ell^{th}$ layer; the distance is measured from the
                   top of the layer.
        field[ell,:] - any field quantity evaluated at x[ell,:]
        WARNING:  The field values and discretization is dropped for 'transmitted' and 
                  'reflected' half-spaces ($\ell=0,p$) during OUTPUT computation, i.e.,
                  x[0,:], field[0,:], x[-1,:], field[-1,:]
                  must exist for INPUT but are NOT used in computation.

        INPUT (optional):
        xOutside[0,:] - discretization of $0^{th}$ region ('transmitted' half-space); the
                        distance is measured from the top of the region, i.e., 1--0 interface.
        xOutside[-1,:] - discretization of $p^{th}$ region ('reflected' half-space); the distance
                         is measured from 'in-region' to the $p$--$p-1$ interface, i.e.,
                         xOutside[-1,-1] is the furthest point from the $p$--$p-1$ interface and
                         xOutside[-1,0] is the closest point to the $p$--$p-1$ interface 
        field[0,:] - any field quantity evaluated at xOutside[0,:]
        field[-1,:] - any field quantity evaluated at xOutside[-1,:]

        OUTPUT:
        x - flat array containing  discretization of distance as measured from the origin
            of the coordinate system (bottom of 'reflected' half-space).
        field - flat array containing values of the transmitted field at x.
        '''
        # Make local copy of all input arguments, because they will be modified.
        x = x.copy()
        field = field.copy()
        insideCase = (xOutside is None) and (fieldOutside is None)
        if not insideCase:
            xOutside = xOutside.copy()
            fieldOutside = fieldOutside.copy()
        

        for ell in xrange(1, self.h.size-1):
            # Convert x[ell] to be measured from the origin of the coordinate system instead
            # of the top of each layer [see Fig. 1 in the paper]
            x[ell] = self.h[ell:-1].sum() - self.h[ell] + x[ell]
        
        # Because the origin of our coordinate system is set on the the $p$--$p-1$ interface
        # we must flip all columns in order to obtain a continuous x.  
        # For example, if p=4, h=3 for each layer, and xSample=4, then
        # BEFORE THE FLIP:
        # x[1,:] = [6.0, 7.0, 8.0]
        # x[2,:] = [3.0, 4.0, 5.0]
        # x[3,:] = [0.0, 1.0, 2.0]
        # AFTER THE FLIP:
        # x[1,:] = [0.0, 1.0, 2.0]
        # x[2,:] = [3.0, 4.0, 5.0]
        # x[3,:] = [6.0, 7.0, 8.0]
        # and then we can concatenate the rows to obtain a continuous x.    
        x = sp.flipud(x)
        field = sp.flipud(field)

        # Dropping 'transmitted' and 'reflected' half-spaces, i.e., $\ell=0,p$
        # because they contain all zero values, see function header.
        x = x[1:-1]
        field = field[1:-1]

        # Flatten all arrays and obtain a continuous x
        x = x.reshape(x.size)
        field = field.reshape(field.size)
        
        if not insideCase:
            # Computing distance in 'transmitted' half-space as measured 
            # from the origin of the coordinate system.
            xOutside[0] = self.h.sum() + xOutside[0]

            x = sp.concatenate((xOutside[-1], x, xOutside[0]))
            field = sp.concatenate((fieldOutside[-1], field, fieldOutside[0]))

        return x, field     
#---------------------------------------------------------------------------------------------------
    def field(self, fo, phi, xSample, pol, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes transverse E-field (E_y if 'parallel' polarization or E_z if 'perp' polarization)
        as a function of distance for a given frequency, angle of incidence, and polarization.
        [see (33) and (37) in the paper]

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        xSample - number of sample points per layer
                - scalar, integer
        pol - polarization it can be 'parallel' of 'perp' for perpendicular
            - string
        numAvgLayersOutside - How far (measured in average layer thickness) 
                              into 'reflected' and 'transmitted' half-spaces 
                              should the field by computed.
                            - scalar, integer

        OUTPUT:
        x - discretization of distance as measured from the origin of the coordinate system,
            $p$--$p-1$ interface (bottom of 'reflected' half-space) [see Fig. 1 in the paper]
          - 1D array, float      
        gammaPlus - transverse component of the E-field associated with the wave propagating 
                    in the positive x-direction evaluated at x
                  - 1D array, complex
        gammaMinus - transverse component of the E-field associated with the wave propagating 
                     in the negative x-direction evaluated at x
                   - 1D array, complex

        REMARKS:
        The computed transverse E-field is normalized to $\chi^{+}_p$.  $\chi^{+}_p$ is the value
        of the incident transverse E-field at the $p$--$p-1$ interface. [see Fig. 1 in the paper]

        EXAMPLES:
        x, gammaPlus, gammaMinus = field(100.0, 30.0, 'perp')
        x, gammaPlus, gammaMinus = field(400,0, 45.0, 'parallel', 20)
        '''
        # Compute chiPlus & chiMinus in all layers at a given freq., incident angle, & polarization.
        phi = sp.deg2rad(phi)
        bd.setChi(self, fo, phi, pol)

        num = numAvgLayersOutside
        if num == 0:
            x, gammaPlus, gammaMinus = self.__field(xSample, 0)
            gammaPlus = self.__field2CoordSys(x, gammaPlus)[1]
            x, gammaMinus = self.__field2CoordSys(x, gammaMinus)
        elif num > 0:
            x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside = \
                self.__field(xSample, num)
            gammaPlus = self.__field2CoordSys(x, gammaPlus, xOutside, gammaPlusOutside)[1]
            x, gammaMinus = self.__field2CoordSys(x, gammaMinus, xOutside, gammaMinusOutside)

        return x, gammaPlus, gammaMinus
#---------------------------------------------------------------------------------------------------
    def __energy(self, fo, phi, gammaPlus, gammaMinus, case='inside'):
        '''
        DESCRIPTION:
        Computes electric and magnetic energy densities inside/outside of the multilayer stack.
        [see (35), (36), (39) and (40) in the paper] 

        INPUT:
        fo - frequency in Hz (scalar, float)
        phi - angle of incidence in radians (scalar, float)

        If case = 'inside' then gammaPlus[ell,:] and gammaMinus[ell,:] have exactly the same
        meaning as in __field.

        If case = 'outside' then gammaPlus and gammaMinus have exactly the same meaning as 
        gammaPlusOutside and gammaMinusOutside in __field, respectively.

        OUTPUT:
        If case = 'inside' then  __energy returns
        u_e[ell,:] - electric energy density in $\ell^{th}$ layer at x[ell,:], where x[ell,:] is a 
                     discretization of $\ell^{th}$ layer, measured from the top of the layer.  
        u_m[ell,:] - magnetic energy density in $\ell^{th}$ layer at x[ell,:], where x[ell,:] is a 
                     discretization of $\ell^{th}$ layer, measured from the top of the layer.  
        Note: u_e and u_m contain half-spaces with all values set to zero.
              The reason for doing so, is to keep indexing the same as in the paper, i.e.,
              u_e[0,:], u_m[0,:] refers to 'transmitted' half-space and
              u_e[-1,:], u_m[-1,:] refers to 'reflected' half-space.

        If case = 'outside' then __energy returns
        u_e[0,:] - electric energy density in 'transmitted' half-space at x[0,:], where
                   x[0,:] is a discretization of the 'transmitted' half-space.
        u_e[-1,:] - electric energy density in 'reflected' half-space at x[-1,:], where 
                    x[-1,:] is a discretization of the 'reflected' half-space.
        u_m[0,:] - magnetic energy density in 'transmitted' half-space at x[0,:], where
                   x[0,:] is a discretization of the 'transmitted' half-space.
        u_m[-1,:] - magnetic energy density in 'reflected' half-space at x[-1,:], where 
                    x[-1,:] is a discretization of the 'reflected' half-space. 

        WARNINGS:
        1.  This function should NOT be used if a multilayer stack contains any LHM layers.  
            [see the end of Sec. 6 in the paper]
        2.  Both electric and magnetic energy densities are scaled to the total incident energy 
            density evaluated at the $p$--$p-1$ interface.
        '''
        # Compute epsilon, mu, kx, w
        if case == 'inside':
            eps = self.epsRel[:,sp.newaxis] * constants.epsilon_0
            mu = self.muRel[:,sp.newaxis] * constants.mu_0
            kx = self.kx[:,sp.newaxis]
            w = self.w[:,sp.newaxis]
        elif case == 'outside':
            eps = sp.zeros(2, complex)
            eps[0], eps[-1] = self.epsRel[0], self.epsRel[-1]
            eps = eps[:,sp.newaxis] * constants.epsilon_0

            mu = sp.zeros(2, complex)
            mu[0], mu[-1] = self.muRel[0], self.muRel[-1]
            mu = mu[:,sp.newaxis] * constants.mu_0
            
            kx = sp.zeros(2,complex)
            kx[0], kx[-1] = self.kx[0], self.kx[-1]
            kx = kx[:,sp.newaxis]

            w = sp.zeros(2, complex)
            w[0], w[-1] = self.w[0], self.w[-1]
            w = w[:,sp.newaxis]

        omega = 2.0 * sp.pi * fo
        kypSquared = eps[-1].real * mu[-1].real * omega**2 * sp.sin(phi)**2
        kRatioSquared  = kypSquared / sp.absolute(kx)**2

        # computation of the electric and magnetic energy densities
        u_e = sp.zeros(gammaPlus.shape, complex)
        u_m = sp.zeros(gammaPlus.shape, complex)
        if self.pol == 'parallel':
            # normalize energy density to the TOTAL incident energy density, hence the factor of 2
            norm = 2.0 * eps[-1].real/4.0 * 1.0/sp.cos(phi)**2

            # compute electric and magnetic energy densities [see (35) and (36) in the paper]
            u_e = eps.real/4.0 * (
                (1.0 + kRatioSquared) * (sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2)
                + 2.0 * (1.0 - kRatioSquared) * sp.real(gammaPlus * gammaMinus.conj()) )

            u_m = mu.real/4.0 * sp.absolute(w)**2 * (
                sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2 
                - 2.0 * sp.real(gammaPlus * gammaMinus.conj()) )
        elif self.pol == 'perp':
            # normalize energy density to the TOTAL incident energy density, hence the factor of 2
            norm = eps[-1].real/4.0

            # compute electric and magnetic energy densities [see (39) and (40) in the paper]
            u_e = eps.real/4.0 * ( sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2
                                   + 2.0 * sp.real(gammaPlus * gammaMinus.conj()) )

            u_m = mu.real/4.0 * sp.absolute(w)**2 * (
                (1.0 + kRatioSquared) * (sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2)
                - 2.0 * (1.0 - kRatioSquared) * sp.real(gammaPlus * gammaMinus.conj()) )

        return u_e.real/norm, u_m.real/norm
#---------------------------------------------------------------------------------------------------
    def energy(self, fo, phi, xSample, pol, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes electric and magnetic energy densities as a function of distance at a given 
        frequency, angle of incidence, and polarization. [see (35), (36), (39) and (40) in the paper]

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        xSample - number of sample points per layer
                - scalar, integer
        pol - polarization it can be 'parallel' or 'perp' for perpendicular
            - string
        numAvgLayersOutside - How far (measured in average layer thickness) 
                              into 'reflected' and 'transmitted' half-infinite spaces
                              should the field by computed.
                            - scalar, integer

        OUTPUT:
        x - discretization of distance as measured from the origin of the coordinate system,
            $p$--$p-1$ interface (bottom of 'reflected' half-space) [see Fig. 1 in the paper]
          - 1D array, float  
        u_e - contains values of the electric energy density at x
            - 1D array, float
        u_m - contains values of the magnetic energy density at x
            - 1D array, float

        WARNINGS:
        1.  This function should NOT be used if a multilayer stack contains any LHM layers.  
            [see the end of Sec. 6 in the paper]
        2.  Both electric and magnetic energy densities are scaled to the total incident energy 
            density evaluated at the $p$--$p-1$ interface.

        EXAMPLES:
        x, u_e, u_m = energy(100.0, 30.0, 'perp')
        x, u_e, u_m = energy(400,0, 45.0, 'parallel', 20)
        '''
        phi = sp.deg2rad(phi)
        bd.setChi(self, fo, phi, pol)

        # issue a warning if LHM is detected
        LHM = sp.any(sp.signbit(self.epsRel.real) & sp.signbit(self.muRel.real))
        if LHM: 
            warn(LHMwarning)
            

        num = numAvgLayersOutside
        if num == 0:
            x, gammaPlus, gammaMinus = self.__field(xSample, 0)
            u_e, u_m = self.__energy(fo, phi, gammaPlus, gammaMinus, 'inside')
            
            # convert electric and magnetic energy densities to a function of distance
            u_e = self.__field2CoordSys(x, u_e)[1]
            x, u_m = self.__field2CoordSys(x, u_m)
        elif num > 0:
            x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside = \
                self.__field(xSample, num)
            u_eInside, u_mInside = self.__energy(fo, phi, gammaPlus, gammaMinus, 'inside')
            u_eOutside, u_mOutside = self.__energy(fo, phi, gammaPlusOutside, 
                                                          gammaMinusOutside, 'outside')

            # convert electric and magnetic energy densities to a function of distance
            u_e = self.__field2CoordSys(x, u_eInside, xOutside, u_eOutside)[1]
            x, u_m = self.__field2CoordSys(x, u_mInside, xOutside, u_mOutside)

        return x, u_e, u_m
#---------------------------------------------------------------------------------------------------
    def loss(self, fo, phi, xSample, pol, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes electric and magnetic loss densities as a function of distance at a given 
        frequency, angle of incidence, and polarization.  [see (31d) and (31e) in the paper]

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        xSample - number of sample points per layer
                - scalar, integer
        pol - polarization it can be 'parallel' or 'perp' for perpendicular
            - string
        numAvgLayersOutside - How far (measured in average layer thickness) 
                              into 'reflected' and 'transmitted' half-infinite spaces
                              should the field by computed.
                            - scalar, integer

        OUTPUT:
        x - discretization of distance as measured from the origin of the coordinate system,
            $p$--$p-1$ interface (bottom of 'reflected' half-space) [see Fig. 1 in the paper]
          - 1D array, float  
        Q_e - contains values of the electric loss density at x
            - 1D array, float
        Q_m - contains values of the magnetic loss density at x
            - 1D array, float

        WARNINGS:
        1.  This function should NOT be used if a multilayer stack contains any LHM layers.  
            [see the end of Sec. 6 in the paper]
        2.  Both electric and magnetic loss densities are scaled to the total incident energy 
            density evaluated at the $p$--$p-1$ interface.

        EXAMPLES:
        x, Q_e, Q_m = loss(100.0, 30.0, 'perp')
        x, Q_e, Q_m = loss(400,0, 45.0, 'parallel', 20)
        '''
        phi = sp.deg2rad(phi)
        bd.setChi(self, fo, phi, pol)

        # issue a warning if LHM is detected
        LHM = sp.any(sp.signbit(self.epsRel.real) & sp.signbit(self.muRel.real))
        if LHM: 
            warn(LHMwarning)

        omega = 2.0 * sp.pi * fo
        u_e2Q_e = 2.0 * self.epsRel.imag * omega / self.epsRel.real
        u_m2Q_m = 2.0 * self.muRel.imag * omega / self.muRel.real
            
        num = numAvgLayersOutside
        if num == 0:
            x, gammaPlus, gammaMinus = self.__field(xSample, 0)
            u_e, u_m = self.__energy(fo, phi, gammaPlus, gammaMinus, 'inside')
            Q_e = u_e2Q_e[:,sp.newaxis] * u_e
            Q_m = u_m2Q_m[:,sp.newaxis] * u_m 
            
            # convert electric and magnetic loss densities to a function of distance
            Q_e = self.__field2CoordSys(x, Q_e)[1]
            x, Q_m = self.__field2CoordSys(x, Q_m)
        elif num > 0:
            x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside = \
                self.__field(xSample, num)
            u_eInside, u_mInside = self.__energy(fo, phi, gammaPlus, gammaMinus, 'inside')
            u_eOutside, u_mOutside = self.__energy(fo, phi, gammaPlusOutside, 
                                                          gammaMinusOutside, 'outside')

            Q_eOutside = sp.zeros(u_eOutside.shape, float)
            Q_mOutside = sp.zeros(u_mOutside.shape, float)
            Q_eOutside[0], Q_eOutside[-1] = u_e2Q_e[0] * u_eOutside[0], u_e2Q_e[-1] * u_eOutside[-1]
            Q_mOutside[0], Q_mOutside[-1] = u_m2Q_m[0] * u_mOutside[0], u_m2Q_m[-1] * u_mOutside[-1]
            Q_eInside = u_e2Q_e[:,sp.newaxis] * u_eInside
            Q_mInside = u_m2Q_m[:,sp.newaxis] * u_mInside

            # convert electric and magnetic loss densities to a function of distance
            Q_e = self.__field2CoordSys(x, Q_eInside, xOutside, Q_eOutside)[1]
            x, Q_m = self.__field2CoordSys(x, Q_mInside, xOutside, Q_mOutside)

        return x, Q_e, Q_m
#---------------------------------------------------------------------------------------------------
    def __divS(self, fo, phi, gammaPlus, gammaMinus, case='inside'):
        '''
        DESCRIPTION:
        Computes divergence of the time-average complex Poynting vector inside/outside 
        of the multilayer stack.  [see Sec. 6 in the paper]

        INPUT:
        fo - frequency in Hz (scalar, float)
        phi - angle of incidence radians (scalar, float)

        If case = 'inside' then gammaPlus[ell,:] and gammaMinus[ell,:] have exactly the same
        meaning as in __field.

        If case = 'outside' then gammaPlus and gammaMinus have exactly the same meaning as 
        gammaPlusOutside and gammaMinusOutside in __field, respectively.

        OUTPUT:
        If case = 'inside' then  __divS returns
        divS[ell,:] - divergence of the time-average complex Poynting vector in $\ell^{th}$ layer 
                      at x[ell,:], where x[ell,:] is a discretization of $\ell^{th}$ layer, 
                      measured from the top of the layer.  
        Note: divS contains half-spaces with all values set to zero.
              The reason for doing so, is to keep indexing the same as in the paper, i.e.,
              divS[0,:] refers to 'transmitted' half-space and divS[-1,:] to 'reflected' half-space

        If case = 'outside' then __divS returns
        divS[0,:] - divergence of the time-averaged complex Poynting vector in 'transmitted' 
                    half-space at x[0,:], where x[0,:] is a discretization of the 
                    'transmitted' half-space.
        divS[-1,:] - divergence of the time-averaged complex Poynting vector in 'reflected' 
                     half-space at x[-1,:], where x[-1,:] is a discretization of the 
                     'reflected' half-space.


        u_e[0,:] - electric energy density in 'transmitted' half-space at x[0,:], where
                   x[0,:] is a discretization of the 'transmitted' half-space.
        u_e[-1,:] - electric energy density in 'reflected' half-space at x[-1,:], where 
                    x[-1,:] is a discretization of the 'reflected' half-space.
        u_m[0,:] - magnetic energy density in 'transmitted' half-space at x[0,:], where
                   x[0,:] is a discretization of the 'transmitted' half-space.
        u_m[-1,:] - magnetic energy density in 'reflected' half-space at x[-1,:], where 
                    x[-1,:] is a discretization of the 'reflected' half-space. 

        WARNINGS:
        1.  This function should NOT be used if a multilayer stack contains any LHM layers.  
            [see the end of Sec. 6 in the paper]
        2.  $div \cdot \mathbf{S}$ is scaled to the total incident energy density evaluated
            at the $p$--$p-1$ interface.
        '''
        # compute epsilon, mu and kx
        if case == 'inside':
            eps = self.epsRel[:,sp.newaxis] * constants.epsilon_0
            mu = self.muRel[:,sp.newaxis] * constants.mu_0
            kx = self.kx[:,sp.newaxis]
        elif case == 'outside':
            eps = sp.zeros(2, complex)
            eps[0], eps[-1] = self.epsRel[0], self.epsRel[-1]
            eps = eps[:,sp.newaxis] * constants.epsilon_0

            mu = sp.zeros(2, complex)
            mu[0], mu[-1] = self.muRel[0], self.muRel[-1]
            mu = mu[:,sp.newaxis] * constants.mu_0
            
            kx = sp.zeros(2,complex)
            kx[0], kx[-1] = self.kx[0], self.kx[-1]
            kx = kx[:,sp.newaxis]

        omega = 2.0 * sp.pi * fo

        # computation of the divS
        divS = sp.zeros(gammaPlus.shape, complex)
        if self.pol == 'parallel':
            # normalize divS to the total incident energy density, hence the factor of 2.
            norm = 2.0 * eps[-1].real/4.0 * 1.0/sp.cos(phi)**2

            divS = -eps.conj() * omega / kx.conj() * (
                kx.imag * (sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2)
                + 2.0J * kx.real * sp.real(gammaPlus * gammaMinus.conj()) )

        elif self.pol == 'perp':
            # normalize divS to the total incident energy density, hence the factor of 2.
            norm = eps[-1].real/4.0

            divS = -kx.conj() / (mu.conj() * omega) * (
                kx.imag * (sp.absolute(gammaPlus)**2 + sp.absolute(gammaMinus)**2)
                + 2.0J * kx.real * sp.real(gammaPlus * gammaMinus.conj()) )            

        return divS/norm
#---------------------------------------------------------------------------------------------------
    def divPoynting(self, fo, phi, xSample, pol, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes divergence of the time-averaged complex Poynting vector as a function of distance 
        at a given frequency, angle of incidence, and polarization. [see Sec. 6 in the paper]

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        xSample - number of sample points per layer
                - scalar, integer
        pol - polarization it can be 'parallel' or 'perp' for perpendicular
            - string
        numAvgLayersOutside - How far (measured in average layer thickness) 
                              into 'reflected' and 'transmitted' half-infinite spaces
                              should the field by computed.
                            - scalar, integer

        OUTPUT:
        x - discretization of distance as measured from the origin of the coordinate system,
            $p$--$p-1$ interface (bottom of 'reflected' half-space) [see Fig. 1 in the paper]
          - 1D array, float
        divS - contains values of the divergence of the time-average complex Poynting vector at x
             - 1D array, float

        WARNINGS:
        1.  This function should NOT be used if a multilayer stack contains any LHM layers.  
            [see the end of Sec. 6 in the paper]
        2.  $div \cdot \mathbf{S}$ is scaled to the total incident energy density evaluated
            at the $p$--$p-1$ interface.

        EXAMPLES:
        x, divS = divPoynting(100.0, 30.0, 'perp')
        x, divS = divPoynting(400,0, 45.0, 'parallel', 20)
        '''
        phi = sp.deg2rad(phi)
        bd.setChi(self, fo, phi, pol)

        # issue a warning if LHM is detected
        LHM = sp.any(sp.signbit(self.epsRel.real) & sp.signbit(self.muRel.real))
        if LHM: 
            warn(LHMwarning)

        num = numAvgLayersOutside
        if num == 0:
            x, gammaPlus, gammaMinus = self.__field(xSample, 0)
            divS = self.__divS(fo, phi, gammaPlus, gammaMinus, 'inside')
            
            # convert divergence of the Poynting vector as a function of distance
            x, divS = self.__field2CoordSys(x, divS)
        elif num > 0:
            x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside = \
                self.__field(xSample, num)
            divS_Inside = self.__divS(fo, phi, gammaPlus, gammaMinus, 'inside')
            divS_Outside = self.__divS(fo, phi, gammaPlusOutside, gammaMinusOutside, 'outside')


            # convert divergence as a function of distance
            x, divS = self.__field2CoordSys(x, divS_Inside, xOutside, divS_Outside)

        return x, divS
#---------------------------------------------------------------------------------------------------
    def FIM(self, fo, phi, pol):
        '''
        DESCRIPTION: 
        Computes the "Fundamental Invariant of Multilayers" (FIM) at each layer interface
        at a given frequency, angle of incidence, and polarization. [see (29b), (29c) in the paper]

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        pol - polarization it can be 'parallel' of 'perp' for perpendicular
            - string

        OUTPUT:
        FIMbottom[ell] - FIM at the bottom of the $\ell^{th}$ layer (interface 
                         separating $\ell$--$\ell-1$ layers) for $\ell = 1,\ldots,p$.
                       - 1D array, complex
        FIMtop[ell] - FIM at the top of of the $\ell^{th}$ layer (interface 
                      separating $\ell$--$\ell+1$ layers) for $\ell = 0,\ldots,p-1$.
                    - 1D array, complex

        EXAMPLES:
        FIMbottom, FIMtop = FIM(100.0, 30.0, 'perp')
        FIMbottom, FIMtop = FIM(110.0, 10.0, 'parallel')
        '''
        bd.setChi(self, fo, sp.deg2rad(phi), pol)

        if pol == 'perp':
            q = self.kx / self.muRel
        elif pol == 'parallel':
            q = self.epsRel / self.kx

        # FIM at the bottom of $\ell^{th}$ region for $\ell=1,\ldots,p$. [see Fig. 1 in the paper]
        FIMbottom = (self.chiPlus[1:]**2 - self.chiMinus[1:]**2) * q[1:]

        # FIM at the top of $\ell^{th}$ region for $\ell=0,\ldots,p-1$. [see Fig. 1 in the paper]
        ePlusTop = self.chiPlus[0:-1] * sp.exp(-1.0J * self.kx[0:-1] * self.h[0:-1]) 
        eMinusTop = self.chiMinus[0:-1] * sp.exp(1.0J * self.kx[0:-1] * self.h[0:-1])
        FIMtop = (ePlusTop**2 - eMinusTop**2) * q[0:-1]

        return FIMbottom, FIMtop
#---------------------------------------------------------------------------------------------------
    def FIMvsDist(self, fo, phi, xSample, pol, numAvgLayersOutside=0):
        '''
        DESCRIPTION:
        Computes "Fundamental Invariant of Multilayers" (FIM) as a function of distance at
        a given frequency, angle of incidence, and polarization.

        INPUT:
        fo - frequency in Hz
           - scalar, float
        phi - angle of incidence in degrees
            - scalar, float
        xSample - number of sample points per layer
                - scalar, integer
        pol - polarization it can be 'parallel' of 'perp' for perpendicular
            - string
        numAvgLayersOutside - How far (measured in average layer thickness) 
                              into 'reflected' and 'transmitted' half-spaces
                              should the field by computed.
                            - scalar, integer

        OUTPUT:
        x - discretization of distance as measured from the origin of the coordinate system,
            $p$--$p-1$ interface (bottom of 'reflected' half-space) [see Fig. 1 in the paper]
          - 1D array, float
        FIM - value of the "Fundamental Invariant of Multilayers" at x
             - 1D array, complex

        EXAMPLES:
        x, FIM = FIMvsDist(100.0, 30.0, 'perp')
        x, FIM = FIMvsDist(400,0, 45.0, 'parallel', 20)
        '''
        phi = sp.deg2rad(phi)
        bd.setChi(self, fo, phi, pol)

        num = numAvgLayersOutside
        # compute $\gamma^{\pm}(x)$ in each layer and half-spaces if num > 0 
        # [see (33) and (37) in the paper]
        if num == 0:
            x, gammaPlus, gammaMinus = self.__field(xSample, 0)
        elif num > 0:
            x, gammaPlus, gammaMinus, xOutside, gammaPlusOutside, gammaMinusOutside = \
                self.__field(xSample, num)

        # compute scaled $w_{\ell}$ [see (16) and (22) in the paper]
        if pol == 'perp':
            w = self.kx / self.muRel 
            if num > 0:
                wOutside = sp.zeros(2, complex)
                wOutside[0], wOutside[-1] = self.kx[0]/self.muRel[0], self.kx[-1]/self.muRel[-1]
        elif pol == 'parallel':
            w = self.epsRel / self.kx
            if num > 0:
                wOutside = sp.zeros(2, complex)
                wOutside[0], wOutside[-1] = self.epsRel[0]/self.kx[0], self.epsRel[-1]/self.kx[-1]

        # compute FIM inside $\ell = 1,..,p-1$ regions [see (29a) in the paper]
        w = w[:,sp.newaxis]
        FIM = w * (gammaPlus**2 - gammaMinus**2)

        # convert FIM to the coordinate system, i.e., FIM as a function of distance
        if num == 0:
            x, FIM = self.__field2CoordSys(x, FIM)
        elif num > 0:
            # compute FIM for transmitted and reflected half-spaces.
            wOutside = wOutside[:,sp.newaxis]
            FIMoutside = wOutside * (gammaPlusOutside**2 - gammaMinusOutside**2)
            x, FIM = self.__field2CoordSys(x, FIM, xOutside, FIMoutside)
                
        return x, FIM
#---------------------------------------------------------------------------------------------------
    def TRvsFreq(self, fmin, fmax, fSample, phi, pol):
        '''
        DESCRIPTION:
        Computes transmission and reflection coefficients as a function of frequency at a given
        angle of incidence, and polarization, i.e., $T(f)$ and $R(f)$ [see Table 2 in the paper]
        
        INPUT: 
        fmin - minimal frequency in Hertz
             - scalar, float
        fmax - maximum frequency in Hertz
             - scalar, float
        fSample - number of sample points in [fmin,fmax] interval
                - scalar, integer
        phi - angle of incidence in degrees
            - scalar, float
        pol - polarization, 'parallel' or 'perp' for perpendicular
            - string

        OUTPUT: 
        f - frequency in Hertz
          - 1D array, float
        T - transmission coefficient at f, i.e., $T(f)$
          - 1D array, float
        R - reflection coefficient at f, i.e., $R(f)$
          - 1D array, float

        EXAMPLES: 
        Computes T and R in [1Hz,100Hz] frequency interval with 1000 sample points
        at an angle of incidence of 20 degrees with parallel polarization.
        f, T, R = TRvsFreq(1.0, 100.0, 1000, 20.0, 'parallel')
        '''
        phi = sp.deg2rad(phi)
        f = sp.linspace(fmin, fmax, fSample)
        T = sp.zeros(f.size, float)
        R = sp.zeros(f.size, float)
        for ix in xrange(f.size):
            bd.setChi(self, f[ix], phi, pol)
            T[ix] = self.__T()
            R[ix] = self.__R()

        return f, T, R
#---------------------------------------------------------------------------------------------------
    def TRvsAngle(self, phiMin, phiMax, phiSample, f, pol):
        '''
        DESCRIPTION:
        Computes transmission and reflection coefficients as a function of angle of incidence
        at a given frequency, and polarization, i.e., $T(\phi)$ and $R(\phi)$, where $\phi$ is
        in degrees.  [see Table 2 in the paper]

        INPUT: 
        phiMin - minimal angle of incidence in degrees
               - scalar, float
        phiMax - maximum angle of incidence in degrees
               - scalar, float
        phiSample - number of sample points in [phiMin,phiMax] interval
                  - scalar, integer
        f - frequency in Hertz
          - scalar, float
        pol - polarization, 'parallel' or 'perp' for perpendicular
            - string

        OUTPUT: 
        phi - angle of incidence in degrees
            - 1D array, float
        T - transmission coefficient at phi, i.e., $T(\phi)$
          - 1D array, float
        R - reflection coefficient at phi, i.e., $R(\phi)$
          - 1D array, float

        EXAMPLES:
        Computes T and R in [10 degrees, 30 degrees] incidence angle interval with 100 sample points
        at a frequency of 150.0 Hz with perpendicular polarization.
        phi, T, R = TRvsAngle(10.0, 30.0, 100, 150.0, 'perp')
        '''
        phi = sp.deg2rad( sp.linspace(phiMin, phiMax, phiSample) )
        T = sp.zeros(phi.size, float)
        R = sp.zeros(phi.size, float)
        for ix in xrange(phi.size):
            bd.setChi(self, f, phi[ix], pol)
            T[ix] = self.__T()
            R[ix] = self.__R()

        return sp.rad2deg(phi), T, R
#---------------------------------------------------------------------------------------------------
    def TRvsFreqAndAngle(self, fmin, fmax, fSample, phiMin, phiMax, phiSample, pol):
        '''
        DESCRIPTION:
        Computes transmission and reflection coefficients as a function of frequency and angle of
        incidence at a given polarization, i.e, $T(f,\phi)$ and $T(f,\phi)$.  
        [see Table 1 in the paper]
        
        INPUT: 
        fmin - minimal frequency in Hertz
             - scalar, float
        fmax - maximum frequency in Hertz
             - scalar, float
        fSample - number of sample points in [fmin,fmax] interval
                - scalar, integer
        phiMin - minimal angle of incidence in degrees
               - scalar, float
        phiMax - maximum angle of incidence in degrees
               - scalar, float
        phiSample - number of sample points in [phiMin,phiMax] interval
                  - scalar, integer
        pol - polarization, 'parallel' or 'perp' for perpendicular
            - string

        OUTPUT: 
        fMat - frequency in Hertz; fMat[i,:] = sp.linspace(fmin, fmax, fSample)
             - 2D array, float
        phiMat - angle of incidence in degrees; phiMat[:,i] = sp.linspace(phiMin, phiMax, phiSample)
               - 2D array, float
        T - transmission coefficient at fMat and phiMat, i.e., $T(f,\phi)$
          - 2D array, float
        R - reflection coefficient at fMat and phiMat, i.e., $R(f,\phi)$
          - 2D array, float

        EXAMPLES: 
        Computes T and R in [1Hz, 100Hz] frequency interval with 1000 sample points
        and [10 degrees, 40 degrees] angle of incidence interval with 200 sample points  
        (parallel polarization).
        f, phi, T, R = TRvsFreqAndAngle(1.0, 100.0, 1000, 10.0, 40.0, 200, 'parallel')
        '''
        # Discretize frequency and angle of incidence
        fVec = sp.linspace(fmin, fmax, fSample)
        phiVec = sp.deg2rad( sp.linspace(phiMin, phiMax, phiSample) )

        # Each row of fMat contains fVec and each column of phiMat contains phiVec
        fMat, phiMat = sp.meshgrid(fVec, phiVec)
        Nrows, Ncols = fMat.shape

        # Each row of T/R contains T/R for all frequencies at a fixed angle on incidence.
        # Each column of T/R contains T/R for all angles of incidence at fixed frequency.
        T = sp.zeros(fMat.shape, float)
        R = sp.zeros(fMat.shape, float)
        for row in xrange(Nrows):
            for col in xrange(Ncols):
                bd.setChi(self, fMat[row,col], phiMat[row,col], pol)
                R[row,col] = self.__R()
                T[row,col] = self.__T()

        return fMat, sp.rad2deg(phiMat), T, R
#---------------------------------------------------------------------------------------------------
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MAY BE INCLUDE PHASE CALCULATIONS IN THE NEXT RELEASE % 
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#     def phaseVSfreq(self, fmin, fmax, fSample, phi, pol, wrapped=False):
#         '''
#         DESCRIPTION:  
#         Computes phase of the transverse component (E_y if 'parallel' polarization or
#         E_z if 'perp' polarization) of the transmitted/reflected E-field as a function of frequency
#         at a given angle of incidence, and polarization.

#         INPUT:
#         fmin - minimal frequency in Hertz
#              - scalar, float
#         fmax - maximum frequency in Hertz
#              - scalar, float
#         fSample - number of sample points in [fmin,fmax] interval
#                 - scalar, integer
#         phi - angle of incidence in degrees
#             - scalar, float
#         pol - polarization, 'parallel' or 'perp' for perpendicular
#             - string
#         wrapped - (optional) If set to True then wrapped phase is returned.  
#                   If set to False (default) then unwrapped phase is returned.
#                 - boolean

#         OUTPUT:
#         f - frequency in Hertz
#           - 1D array, float
#         Tphase - transmitted E-field phase in degrees at f
#                - 1D array, float
#         Rphase - reflected E-field phase in degrees at f
#                - 1D array, float

#         REMARKS:
#         The computed phase is relative to the phase of the incident transverse E-field at the
#         interface between the 'reflected' half-infinite space and the $p-1$ layer.  In other words,
#         the computed phase is phase of $\frac{\chi^{\pm}_0}{\chi^{+}_p}$.

#         WARNING:
#         Only the phase of the transverse E-field is computed.  It's not even clear what one
#         means by phase at non-zero angle of incidence with parallel polarization because E-field
#         has two components.

#         EXAMPLES:
#         f, Tphase, Rphase = phaseVSfreq(10.0, 100.0, 1000, 30, 'perp')
#         f, Tphase, Rphase = phaseVSfreq(20.0, 200.0, 2000, 60, 'parallel', True)
#         '''
#         phi = sp.deg2rad(phi)
#         f = sp.linspace(fmin, fmax, fSample)

#         # Compute wrapped phase for each frequency
#         Tphase = sp.zeros(f.shape, float); Rphase = sp.zeros(f.shape, float)
#         for ix in xrange(f.size):
#             bd.setChi(self, f[ix], phi, pol)
#             Tphase[ix] = sp.angle(self.chiPlus[0])
#             Rphase[ix] = sp.angle(self.chiMinus[-1])

#         # Unwrap phase
#         if not wrapped:
#             Tphase = sp.unwrap(Tphase)
#             Rphase = sp.unwrap(Rphase)

#         return f, sp.rad2deg(Tphase), sp.rad2deg(Rphase)
# #---------------------------------------------------------------------------------------------------
#     def phaseVSangle(self, phiMin, phiMax, phiSample, f, pol, wrapped=False):
#         '''
#         DESCRIPTION:  
#         Computes the phase of the transverse component (E_y if 'parallel' polarization or
#         E_z if 'perp' polarization) of the transmitted/reflected E-field as a function of 
#         angle of incidence at a given frequency, and polarization.

#         INPUT:
#         phiMin - minimal angle of incidence in degrees
#                - scalar, float
#         phiMax - maximum angle of incidence in degrees
#                - scalar, float
#         phiSample - number of sample points in [phiMin,phiMax] interval
#                   - scalar, integer
#         f - frequency in Hertz
#           - scalar, float
#         pol - polarization, 'parallel' or 'perp' for perpendicular
#             - string
#         wrapped - (optional) If set to True then wrapped phase is returned.  
#                   If set to False (default) then unwrapped phase is returned.
#                 - boolean

#         OUTPUT:
#         phi - angle of incidence in degrees
#             - 1D array, float
#         Tphase - transmitted E-field phase in degrees at phi
#                - 1D array, float
#         Rphase - reflected E-field phase in degrees at phi
#                - 1D array, float

#         REMARKS:
#         The computed phase is relative to the phase of the incident transverse E-field at the
#         interface between the 'reflected' half-infinite space and the $p-1$ layer.  In other words,
#         the computed phase is phase of $\frac{\chi^{\pm}_0}{\chi^{+}_p}$.

#         WARNING:
#         Only the phase of the transverse E-field is computed.  It's not even clear what one
#         means by phase at non-zero angle of incidence with parallel polarization because E-field
#         has two components.

#         EXAMPLES:
#         phi, Tphase, Rphase = phaseVSfreq(10.0, 100.0, 1000, 30, 'perp')
#         phi, Tphase, Rphase = phaseVSfreq(20.0, 200.0, 2000, 60, 'parallel', True)
#         '''
#         phi = sp.deg2rad( sp.linspace(phiMin, phiMax, phiSample) )
        
#         # Compute wrapped phase for each angle of incidence
#         Tphase = sp.zeros(phi.shape, float); Rphase = sp.zeros(phi.shape, float)
#         for ix in xrange(phi.size):
#             bd.setChi(self, f, phi[ix], pol)
#             Tphase[ix] = sp.angle(self.chiPlus[0])
#             Rphase[ix] = sp.angle(self.chiMinus[-1])

#         # Unwrap phase
#         if not wrapped:
#             Tphase = sp.unwrap(Tphase)
#             Rphase = sp.unwrap(Rphase)

#         return sp.rad2deg(phi), sp.rad2deg(Tphase), sp.rad2deg(Rphase)
# #---------------------------------------------------------------------------------------------------



import scipy as sp
import openTMM.tests.logFile as LF

#---------------------------------------------------------------------------------------------------
def randomIdealLayers(epsRelRange, muRelRange, hRange, numLayers):
    sp.random.seed()

    epsRel = sp.random.uniform(epsRelRange[0], epsRelRange[1], numLayers)
    muRel = sp.random.uniform(muRelRange[0], muRelRange[1], numLayers)
    h = sp.random.uniform(hRange[0], hRange[1], numLayers)
    
    layersDict = {'height':h, 'epsilonRelative':epsRel, 'muRelative':muRel}
    return layersDict
#---------------------------------------------------------------------------------------------------
def randomLayers(epsRelRange, muRelRange, hRange, numLayers):
    layersDictRe = randomIdealLayers(epsRelRange.real, muRelRange.real, hRange, numLayers)
    layersDictIm = randomIdealLayers(epsRelRange.imag, muRelRange.imag, hRange, numLayers)

    epsRel = layersDictRe['epsilonRelative'] + 1.0J*layersDictIm['epsilonRelative']
    muRel = layersDictRe['muRelative'] + 1.0J*layersDictIm['muRelative']
    h = layersDictRe['height']
    
    layersDict = {'height':h, 'epsilonRelative':epsRel, 'muRelative':muRel}
    return layersDict
#--------------------------------------------------------------------------------------------------- 
def shuffleLayers(layersDict1, layersDict2):
    epsRel = sp.concatenate((layersDict1['epsilonRelative'], layersDict2['epsilonRelative']))
    muRel = sp.concatenate((layersDict1['muRelative'], layersDict2['muRelative']))
    h = sp.concatenate((layersDict1['height'], layersDict2['height']))

    # Shuffle the layers
    ix = range(epsRel.size)
    sp.random.shuffle(ix)
    epsRel = epsRel[ix]
    muRel = muRel[ix]
    h = h[ix]

    layersDict = {'height':h, 'epsilonRelative':epsRel, 'muRelative':muRel}
    return layersDict
#---------------------------------------------------------------------------------------------------
def assertApproxEqual(a, b):
    a = sp.asarray(a)
    b = sp.asarray(b)

    # D. E. Knuth, The Art of Computer Programming, Addison-Wesley, 2nd edition, 1981
    # page. 219, eqn. (34), says
    # $|u - v| \leq \epsilon |u|$ and $|u - v| \leq \epsilon |v|$ implies $ u \approx v $
    # Single precision, http://en.wikipedia.org/wiki/IEEE_754-2008
    rtol = 1.0*10**-7
    atol = 1.0*10**-38
    abSame = sp.allclose(a, b, rtol, atol) and sp.allclose(b, a, rtol, atol)
    return abSame
#---------------------------------------------------------------------------------------------------



import scipy as sp
import time
import pickle
import os
import os.path 

class LogFile:
    def __init__(self, filename=None):
        # Create file if no filename is given.  If filename is given, read and save content.
        if filename is None:
            ct = time.localtime(time.time())
            date = str(ct.tm_mon) + '-' + str(ct.tm_mday) + '-' + str(ct.tm_year)
            t = str(ct.tm_hour) + ':' + str(ct.tm_min) + ':' + str(ct.tm_sec)
        
            self.filename = 'openTMMlog--' + date + '--' + t + '.pkl'
            self.saveFile = False
        elif filename is not None:
            self.filename = filename
            self.saveFile = False
            self.data = self.read()
#---------------------------------------------------------------------------------------------------
    def write(self, data, saveFile=True):
        # Using binary mode
        output = open(self.filename, 'ab')
        pickle.dump(data, output, protocol=2)
        output.close()
        self.saveFile = saveFile
#---------------------------------------------------------------------------------------------------
    def read(self):
        f = open(self.filename, 'rb')
        d = {}; i = 0

        cond = True
        while cond:
            try:
                d[str(i)] = pickle.load(f)
                i += 1
            except EOFError:
                f.close()
                cond = False
        return d
#---------------------------------------------------------------------------------------------------
    def cleanUp(self):
        if not self.saveFile and os.path.isfile(self.filename):
            try:
                os.remove(self.filename)
            except OSError:
                print "Could NOT delete layerStack log file, " + self.filename
#---------------------------------------------------------------------------------------------------
    def minDataSet(self, layerStack, layerDict, data, testName, msg=None):

        def criticalAngles(epsRel, muRel):
            ixIdeal = layerStack._ixIdealLHM | layerStack._ixIdealRHM
            angles = sp.arcsin( sp.sqrt(epsRel[ixIdeal].real * muRel[ixIdeal].real / 
                                        (epsRel[-1].real * muRel[-1].real)) )
            angles *= 180.0/sp.pi
            ix = sp.isreal(angles)
            criticalAngles = angles[ix].real
            return sp.sort(criticalAngles)

        ls = layerStack
        dataSet = {'h':ls.h, 'pol':ls.pol, 'patholMethod':ls._patholMethod, 
                   'data':data, 'testName': testName}

        ld = layerDict
        epsRel = sp.ones(ld['epsilonRelative'].size + 2, complex)
        muRel = sp.ones(ld['muRelative'].size + 2, complex)
        epsRel[1:ld['epsilonRelative'].size+1] = ld['epsilonRelative']
        muRel[1:ld['muRelative'].size+1] = ld['muRelative']
        if ld.has_key('hostMedia'):
            epsRel[0] = ld['hostMedia']['epsilonRelative'][0]
            epsRel[-1] = ld['hostMedia']['epsilonRelative'][1]
            muRel[0] = ld['hostMedia']['muRelative'][0]
            muRel[-1] = ld['hostMedia']['muRelative'][1]

        dataSet['epsRel'] = epsRel
        dataSet['muRel'] = muRel


        dataSet['criticalPhi'] = criticalAngles(ls.epsRel, ls.muRel)
        dataSet['msg'] = msg
        return dataSet
#----------------------------------------------------------------------------------------------------



import scipy as sp
import scipy.constants as constants
import copy
from openTMM import Layer
import openTMM.tests.utils as utils

def test(logfile, layerDict1, layerDict2, freq, phiDeg, energyConserv=False):
    numFailedTests = 0
    numTests = 0

    # freq in Hz (range), phiDeg in Deg (range)
    freqSample = 300 
    phiSample = 100

    freqSet = sp.linspace(freq[0], freq[1], freqSample)
    phiDegSet = sp.linspace(phiDeg[0], phiDeg[1], phiSample)

    for patholMethod in 'keepReal', 'makeComplex':
        layerStack1 = Layer(layerDict1, patholMethod)
        layerStack2 = Layer(layerDict2, patholMethod)

        # Test setChi
        for pol in "parallel", "perp":
            for f in freqSet:
                for phi in sp.deg2rad(phiDegSet):
                    numTests += 1
                    layerStack1.setChi(f, phi, pol)
                    layerStack2.setChi(f, phi, pol)

                    testPassed = utils.assertApproxEqual(sp.absolute(layerStack1.chiPlus[0]), 
                                                         sp.absolute(layerStack2.chiPlus[0]))
                    if not testPassed:
                        data = {'freq': f, 'phi': phi}
                        data['chi'] = {}
                        data['chi']['chi1Plus'] = layerStack1.chiPlus[0]
                        data['chi']['chi2Plus'] = layerStack2.chiPlus[0]

                        dataSet = logfile.minDataSet(layerStack1, layerDict1, data, 'reciprocityChi')
                        logfile.write(dataSet)
                        numFailedTests += 1


        # Test getTRvsFrequencyAndAngle
        for pol in "parallel", "perp":
            numTests += 1
            f1, phi1, T1, R1 = layerStack1.TRvsFreqAndAngle(freq[0], freq[1], freqSample,
                                                            phiDeg[0], phiDeg[1], phiSample, pol)
            f2, phi2, T2, R2 = layerStack2.TRvsFreqAndAngle(freq[0], freq[1], freqSample,
                                                            phiDeg[0], phiDeg[1], phiSample, pol)

            testPassed = utils.assertApproxEqual(T1, T2)
            if not testPassed:
                data = {'freqData':(freq[0], freq[1], freqSample), 
                        'phiData':(phiDeg[0], phiDeg[1], phiSample)} 
                data['T'] = {}
                data['T']['T1'] = T1
                data['T']['T2'] = T2
            
                dataSet = logfile.minDataSet(layerStack1, layerDict1, data, 
                                             'reciprocityTRvsFreqAndAngle')
                logfile.write(dataSet)
                numFailedTests += 1

            if energyConserv:
                numTests += 1
                testPassed = utils.assertApproxEqual(R1, R2)
                if not testPassed:
                    data = {'freqData':(freq[0], freq[1], freqSample), 
                            'phiData':(phiDeg[0], phiDeg[1], phiSample)} 
                    data['R'] = {}
                    data['R']['R1'] = R1
                    data['R']['R2'] = R2
                
                    dataSet = logfile.minDataSet(layerStack1, layerDict1, data, 
                                                 'reciprocityTRvsFreqAndAngle')
                    logfile.write(dataSet)
                    numFailedTests += 1
            
    return numFailedTests, numTests
#---------------------------------------------------------------------------------------------------
def run(logfile):

    def printTestResult(sampleNum, numFailedTests, numTests):
        msg = 'Reciprocity test results for sample ' + str(sampleNum) + ' : '
        if numFailedTests == 0:
            msg = msg + 'Passed ' + str(numTests) + ' Tests'
        elif numFailedTests != 0:
            msg = msg + 'Failed ' + str(numFailedTests) + ' of ' + str(numTests) + ' Tests' 
        print msg

    def printTestStatus(sampleNum):
        print 'Performing reciprocity test on sample ' + str(sampleNum) + ' of 10.'
            

    mm = 1.0*10**-3
    GHz = 1.0*10**9

    freqRange = sp.array([100.0, 150.0], float) * GHz
    phiRange = sp.array([0, 89.0], float)


    # 1st Sample, RHM without absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in mathbb{R^+}$
    # $\epsilon''_{\ell} = 0, \mu''_{\ell} \} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 1
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0, 9.0], float)
    muRange = sp.array([1.0, 9.0], float)
    hRange = sp.array([1, 3], float) * mm

    layerDict1 = utils.randomIdealLayers(epsRange, muRange, hRange, 10)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange, energyConserv=True)
    printTestResult(sampleNum, fNum, total)


    # 2nd Sample, LHM without absorption.
    # \{ $\epsilon'_{\ell}, \mu'_{\ell} \} \in mathbb{R^-}$
    # $\epsilon''_{\ell} = 0, \mu''_{\ell} \} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 2
    printTestStatus(sampleNum)
    layerDict1['epsilonRelative'] *= -1.0
    layerDict1['muRelative'] *= -1.0
    layerDict2['epsilonRelative'] *= -1.0
    layerDict2['muRelative'] *= -1.0

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange, energyConserv=True)
    printTestResult(sampleNum, fNum, total)


    # 3rd Sample, RHM with absorption.
    # \{ $\epsilon'_{\ell}, \mu'_{\ell} \} \in mathbb{R^+}$
    # $\epsilon''_{\ell} = \in \mathbb{R^+}, \mu''_{\ell} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 3
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0 + .01J, 9.0 + .1J], complex)
    muRange = sp.array([1.0, 9.0], float)
    hRange = sp.array([1, 3], float) * mm

    layerDict1 = utils.randomLayers(epsRange, muRange, hRange, 10)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 4th Sample, LHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^-}$
    # $\epsilon''_{\ell} = \in \mathbb{R^+}, \mu''_{\ell} = 0$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 4
    printTestStatus(sampleNum)
    epsRange = sp.array([-1.0 + .01J, -9.0 + .1J], complex)
    muRange = sp.array([-1.0, -9.0], float)
    hRange = sp.array([1, 3], float) * mm

    layerDict1 = utils.randomLayers(epsRange, muRange, hRange, 10)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 5th Sample, RHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^+}$
    # $\{ \epsilon''_{\ell}, \mu''_{\ell} \} \in \mathbb{R^+}$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 5
    printTestStatus(sampleNum)
    epsRange = sp.array([1.0 + .01J, 9.0 + .1J], complex)
    muRange = sp.array([1.0 + .01J, 9.0 + .1J], complex)
    hRange = sp.array([1, 3], float)*mm

    layerDict1 = utils.randomLayers(epsRange, muRange, hRange, 10)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 6th Sample, LHM with absorption.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^-}$
    # $\{ \epsilon''_{\ell}, \mu''_{\ell} \} \in \mathbb{R^+}$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 6
    printTestStatus(sampleNum)
    epsRange = sp.array([-1.0 + .01J, -9.0 + .1J], complex)
    muRange = sp.array([-1.0 + .01J, -9.0 + .1J], complex)
    hRange = sp.array([1, 3], float)*mm

    layerDict1 = utils.randomLayers(epsRange, muRange, hRange, 10)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 7th Sample, RHM and LHM with absorption
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^{\pm}}$
    # $\{ \epsilon''_{\ell}, \mu''_{\ell} \} \in \mathbb{R^+}$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 7
    printTestStatus(sampleNum)
    epsRangeRHM = sp.array([.1 + .01J, 10.0 + 1.0J], complex)
    muRangeRHM = sp.array([.1 + .01J, 10.0 + 1J], complex)
    hRangeRHM = sp.array([.1, 10], float) * mm
    layerDictRHM = utils.randomLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 25)

    epsRangeLHM = sp.array([-.1 + .01J, -10.0 + 1.0J], complex)
    muRangeLHM = sp.array([-.1 + .01J, -10.0 + 1J], complex)
    hRangeLHM = sp.array([.1, 10], float) * mm
    layerDictLHM = utils.randomLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 25)

    layerDict1 = utils.shuffleLayers(layerDictRHM, layerDictLHM)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 8th Sample, RHM and LHM with/without absorption and RHM hostMedia
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^{\pm}}$
    # $\{ \epsilon''_{\ell}, \mu''_{\ell} \} \in \mathbb{R^+} or 0$
    # $\epsilon'_0 = \epsilon_p > 1, \mu'_0 = \mu'_p > 1$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 8
    printTestStatus(sampleNum)
    epsRangeRHM = sp.array([.1 + .01J, 10.0 + 1.0J], complex)
    muRangeRHM = sp.array([.1 + .01J, 10.0 + 1J], complex)
    hRangeRHM = sp.array([.1, 1.0], float) * mm
    layerDictRHM = utils.randomLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 25)

    epsRangeRHMideal = sp.array([.1, 10.0], float)
    muRangeRHMideal = sp.array([.1, 10.0], float)
    hRangeRHMideal = sp.array([.1, 10.0], float) * mm
    layerDictRHMideal = utils.randomIdealLayers(epsRangeRHMideal, muRangeRHMideal, 
                                                hRangeRHMideal, 25)

    epsRangeLHM = sp.array([-.1 + .01J, -10.0 + 1.0J], complex)
    muRangeLHM = sp.array([-.1 + .01J, -10.0 + 1J], complex)
    hRangeLHM = sp.array([.1, 1.0], float) * mm
    layerDictLHM = utils.randomLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 25)

    epsRangeLHMideal = sp.array([-.1, -5.0], float)
    muRangeLHMideal = sp.array([-.1, -3.0], float)
    hRangeLHMideal = sp.array([.1, 2.0], float) * mm
    layerDictLHMideal = utils.randomIdealLayers(epsRangeLHMideal, muRangeLHMideal, 
                                                 hRangeLHMideal, 25)

    layerDictWithAbsorption = utils.shuffleLayers(layerDictRHM, layerDictLHM)
    layerDictWithoutAbsorption = utils.shuffleLayers(layerDictRHMideal, layerDictLHMideal)
    layerDict1 = utils.shuffleLayers(layerDictWithAbsorption, layerDictWithoutAbsorption)
    layerDict1['hostMedia'] = {}
    layerDict1['hostMedia']['epsilonRelative'] = 2.0 * sp.ones(2, float)
    layerDict1['hostMedia']['muRelative'] = 3.0 * sp.ones(2, float)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)


    # 9th Sample, RHM and LHM without absorption and LHM hostMedia.
    # $\{ \epsilon'_{\ell}, \mu'_{\ell} \} \in \mathbb{R^{\pm}}$
    # $\epsilon''_{\ell} = 0, \mu''_{\ell} \} = 0$
    # $\epsilon'_0 = \epsilon_p < -1, \mu'_0 = \mu'_p < -1$
    # -----------------------------------------------------------------------------------------------
    sampleNum = 9
    printTestStatus(sampleNum)
    MHz = 1.0*10**6
    freqRange = sp.array([100.0, 150.0], float) * MHz
    period = constants.c/freqRange[0]

    epsRangeRHM = sp.array([1.0, 5.0], float)
    muRangeRHM = sp.array([1.0, 5.0], float)
    hRangeRHM = sp.array([.1, 1.0], float) * period
    layerDictRHM = utils.randomIdealLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 25)

    epsRangeLHM = sp.array([-1, -5], float)
    muRangeLHM = sp.array([-1, -5], float)
    hRangeLHM = sp.array([.1, 1.0], float) * period
    layerDictLHM = utils.randomIdealLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 25)

    layerDict1 = utils.shuffleLayers(layerDictRHM, layerDictLHM)
    layerDict1['hostMedia'] = {}
    layerDict1['hostMedia']['epsilonRelative'] = -3.0 * sp.ones(2, float)
    layerDict1['hostMedia']['muRelative'] = -1.5 * sp.ones(2, float)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange, energyConserv=True)
    printTestResult(sampleNum, fNum, total)


    # 10th Sample, RHM and LHM with absorption and LHM hostMedia.
    # $ 0 < \{ \epsilon'_{\ell}, \mu'_{\ell} \} < 1$ or 
    # $ -1 < \{ \epsilon'_{\ell}, \mu'_{\ell} \} < 0$
    # $\{ \epsilon''_{\ell} = 0, \mu''_{\ell} \} \in \mathbb{R^+}$
    # $\epsilon'_0 = \epsilon_p < -1, \mu'_0 = \mu'_p < -1$
    #-----------------------------------------------------------------------------------------------
    sampleNum = 10
    printTestStatus(sampleNum)
    MHz = 1.0*10**6
    freqRange = sp.array([100.0, 150.0], float) * MHz
    period = constants.c/freqRange[0]

    epsRangeRHM = sp.array([.01 + .1J, .9 + 1.0J], complex)
    muRangeRHM = sp.array([.01 + .01J, .9 + .1J], complex)
    hRangeRHM = sp.array([.1, 1], float) * period
    layerDictRHM = utils.randomLayers(epsRangeRHM, muRangeRHM, hRangeRHM, 25)

    epsRangeLHM = sp.array([-.9 + .1J, -.01 + 1.0J], complex)
    muRangeLHM = sp.array([-.9 + .01J, -.1 + .1J], complex)
    hRangeLHM = sp.array([.1, 1], float) * period
    layerDictLHM = utils.randomLayers(epsRangeLHM, muRangeLHM, hRangeLHM, 25)

    layerDict1 = utils.shuffleLayers(layerDictRHM, layerDictLHM)
    layerDict1['hostMedia'] = {}
    layerDict1['hostMedia']['epsilonRelative'] = -1.1 * sp.ones(2, float)
    layerDict1['hostMedia']['muRelative'] = -1.3 * sp.ones(2, float)

    layerDict2 = copy.deepcopy(layerDict1)
    layerDict2['epsilonRelative'] = layerDict2['epsilonRelative'][::-1]
    layerDict2['muRelative'] = layerDict2['muRelative'][::-1]
    layerDict2['height'] = layerDict2['height'][::-1]

    fNum, total = test(logfile, layerDict1, layerDict2, freqRange, phiRange)
    printTestResult(sampleNum, fNum, total)
#---------------------------------------------------------------------------------------------------
